#include <alloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dir.h>
#include "protocol.h"
#include "define.h"
#include "node.h"
#include "buffers.h"
#include "hops.h"
#include "tools.h"
#include "symbol.h"
#include "users.h"
#include "dx.h"
#include "announce.h"
#include "clulink.h"
#include "switch.h"
#include "streams.h"
#include "ping.h"
#include "wwv.h"
#include "mail.h"
#include "message.h"
#include "database.h"
#include "pavillon.h"
#include "set.h"
#include "log.h"
#include "coninit.h"
#include "commands.h"
#include "filter.h"
#include "locator.h"

#define	 SPECIAL_DX_

char	PROTOCOL_linkType[65];
char	PROTOCOL_initDone[65];	//TRUE lorsque l'initialisation est termine
int 	PROTOCOL_adjacentVersion[65];
char	PROTOCOL_badCRC[65];
char	PROTOCOL_mailEveryNLines[65];
char	PROTOCOL_mailLinesCounter[65];
unsigned long PROTOCOL_lFwdMesNum[MAX_STREAMS];

int	iProtNum;		//Numro de version (poid fort = n de version MAJEUR, poid faible = n de version mineur)

extern	char	STREAMS_level[MAX_STREAMS];
extern  char	STREAMS_callsign[MAX_STREAMS][10];
extern  int 	FirstStream;
extern	int 	LastStream;
extern	char	SCRIPT_fileName[65][9];
extern	unsigned long 	MAIL_lFwdMsgNumber[MAX_STREAMS];

unsigned int	iCluLinkPos;

int	PROTOCOL_nFixedInit = TRUE;	/* TRUE si on utilise la version corrigee de l'init de connexion */

static char* pszError[] = {"Ok",
			   "Bad callsign",
			   "Bad number of arguments",
			   "Unknown user",
			   "Unknown cluster",
			   "Bad frequency",
			   "Unknown error",
			   "Bad message number",
			   "Bad user count",
			   "String too long",
			   "Bad size",
                           "Bad CRC",
			   "Unknown protocol"};

//-------------------------------------------------------------------------
//------ Envoyer un protocole  tous les clusters -------------------------
//-------------------------------------------------------------------------
void PROTOCOL_sendProt(int StreamNum, int Pc, char * Pavillon, char * CluLink, int Hops, int CluLinkSize, int ToStream, int nWAZ)
{
  int 	index;
  int  	iHops;

  int  	iFstream,
	iLstream;

  char	sHops[5];

  /* Enlever 1 de CluLinkSize (octet 0, qui ne compte pas) */
  CluLinkSize--;

  /* A qui faut-il envoyer ce protocole ? */
  if( ToStream == PROTOCOL_ALL )
  {
	/* Envoyer a tous les clusters */
	iFstream = FirstStream;
	iLstream = LastStream;
  }
  else
  {
	/* Envoyer uniquement sur le stream designe */
	iFstream = iLstream = ToStream;
  }


  /* Balayage des streams, a la recherche de clusters */
  for(index = iFstream; index <= iLstream; index++)
  {
	if( StreamNum == index )
		continue;	/* Pas a soi meme ! */

	if( ! (STREAMS_level[index] & LEVEL_cluster ) )
		continue;	/* Ce n'est pas un cluster */

	if( HOPS_fix(index, PCAS_USER, IN, 100) )
		continue;	/* Mode AsUser */

	if( PROTOCOL_initDone[index] != PROTOCOL_INIT_DONE &&
	    (Pc == 16 || Pc == 17 || Pc == 19 || Pc == 21) )
		continue; 	/* L'init n'est pas terminee */

	/* Protocole valide ? */
	iHops = HOPS_fix(index, Pc, HOPS_OUT, Hops);
	if( iHops < 1 || iHops > 99 )
		continue;	/* Le protocole n'est pas valide */

	/* Filtre sortant */
	if( ! FILTER_isOK(index, FILTER_OUT, nWAZ) )
		continue;	/* Protocole filtre */

	/* Pavillon */
	if( PROTOCOL_linkType[index] == PAVILLON )
	{
		/* Y a t-il un hops ? */
		if( Pc == 10 || Pc == 25 || Pc == 26 || Pc == 27 || (Pc >= 28 && Pc <= 38) || Pc == 40 || (Pc >= 42 && Pc <= 48) || Pc == 51 )
			strcpy(sHops, "");		/* Pas de hop count */
		else
			sprintf(sHops, "H%d^", iHops);	/* Oui, y'a un hops */

		/* Faut-il un tild (~) ? */
		if( (Pc >= 10 && Pc <= 12) ||  Pc == 15 || Pc == 18  || Pc == 23 ||
		    (Pc >= 26 && Pc <= 29) || (Pc >= 34 && Pc <= 38) || Pc == 41 ||
		     Pc == 45 )
			BUFFERS_printBuff(index, OUT, "%s%s~\n", Pavillon, sHops);
		else
			BUFFERS_printBuff(index, OUT, "%s%s\n", Pavillon, sHops);
	}
	else
	/* CluLink */
	{
		if( CluLinkSize == 0 )	/* Par securite */
			break;

		/* Hop count */
		CluLink[4] = (unsigned char) iHops;

		/* Longueur du protocole */
		 CLULINK_putSize(CluLink, CluLinkSize - 1);

		/* Rechercher les protocoles NON BROADCASTES */
		if( Pc != PC10 )
			CluLink[0] |= CL_BROADCAST; /* Mettre a 1 le bit BROADCAST */

		BUFFERS_addBuff_sized(index, CluLink, OUT, CluLinkSize + 1);
	}
  }/* End FOR */
}

//-------------------------------------------------------------------------
//------ Aiguiller le dcodage de la commande reue -----------------------
//-------------------------------------------------------------------------
void PROTOCOL_do(int StreamNum, char * Cmd)
{
  //------ Quel protocol est utilis ? ------
  //Au dpart, le protocol par dfaut est Pavillon software
  //Un entte [MZN-3.ALPHA-$1.0], transmis lors de la phase
  //d'initialisation, permet de commuter vers le protocol CluLink.

  if( PROTOCOL_linkType[StreamNum] == CLU_LINK )
    PROTOCOL_cluLink(StreamNum, Cmd);
  else
    PROTOCOL_pavillon(StreamNum, Cmd);
}

//-------------------------------------------------------------------------
//------ Extraire les arguments -------------------------------------------
//-------------------------------------------------------------------------
int PROTOCOL_extractPC(char* pszString, char** pArg)
{
	int index;
	int nNbArg;
	char* pPtr;

	pPtr= pszString + 5; /* Pointer juste apres le numro de protocole */

	for(index = 0; index < PPARAM; index++)
	{
		pArg[index] = pPtr;

		/* Rechercher le champs suivant */
		while( *pPtr != SNULL && *pPtr != '^' )
			pPtr++;

		/* Fin de la ligne de commande atteinte ? */
		if( *pPtr == SNULL || *pPtr == '~' )
			break;

		*pPtr++ = SNULL;
	}

	/* Sauvegarder le nombre de parametres recus dans le protocole */
	nNbArg = index;

	/* Completer les champs restant par des NULLs */
	for(;index < PPARAM; index++)
		pArg[index] = NULL;

	return nNbArg;
}

//-------------------------------------------------------------------------
//------ Gestion des erreurs de protocole ---------------------------------
//-------------------------------------------------------------------------
void PROTOCOL_errorPC(int StreamNum, int nError, char* pszProtocol)
{
	extern	char * 	PARAMS_logs;

	if( ! SET_check(PARAMS_logs, LOG_PAVILLON_STR) )
		return;

	if( nError > 0 && nError < PROTOCOL_UNKNOWN )
	{
		char	szPcCall[16];

		/* Quel est le call du cluster ? */
		NODE_getNodeCall(StreamNum, 0, szPcCall);

		LOG_write(LOG_PAVILLON_FILE, LOG_DATE_YES,
			  "%s (adjacent : %s)\n%s\n---\n",
			  pszError[nError], szPcCall,
			  pszProtocol);
	}
}

//---------------------------------------------------------------------------
//------ Coonection en mode utilisateur -------------------------------------
//---------------------------------------------------------------------------
void PROTOCOL_asUser(int StreamNum, char * pszString)
{
	char 	szLogger[20];
	char 	szFrequency[20];
	char 	szDxCall[30];
	char 	szComments[80];
	char 	szTime[20];
	char 	szDate[20];
	char 	szServer[20];
	int	index;
	char*	pPtr1;


	if( strlen (pszString) > 255 )
		return;

	/* Identifier un spot DX */
	if( pszString[0] == 'D' && pszString[1] == 'X' && pszString[2] == ' ')
	{
		pPtr1 = pszString + 6;	/* Pointer sur l'indicatif de l'om */

		/* Lire l'indicatif */
		for(index = 0; index < 19; index++)
		{
			if( *pPtr1 != ':' && *pPtr1 != ' ' )
				szLogger[index] = *pPtr1++;
			else
				break;
		}
		szLogger[index] = SNULL;
		strupr(szLogger);

		/* Passer les espaces */
		pPtr1++;
		while( *pPtr1 == ' ' )
			pPtr1++;

		/* Lire la frequence */
		for(index = 0; index < 19; index++)
		{
			if( *pPtr1 != ' ' )
				szFrequency[index] = *pPtr1++;
			else
				break;
		}
		szFrequency[index] = SNULL;

		/* Passer les espaces */
		while( *pPtr1 == ' ' )
			pPtr1++;

		/* Lire l'indicatif du DX */
		for(index = 0; index < 29; index++)
		{
			if( *pPtr1 != ' ' )
				szDxCall[index] = *pPtr1++;
			else
				break;
		}
		szDxCall[index] = SNULL;

		/* Passer les espaces */
		while( *pPtr1 == ' ' )
			pPtr1++;

		/* Lire le commentaire */
		strcpy(szComments, pPtr1);

		/* y avait-il un commentaire ? */
		if( strlen(szComments) < 8 )
		{
			/* Non ... */
			strcpy(szTime, szComments);
			strcpy(szComments, " ");
		}
		else
		{
			/* Oui ... Lire le commentaire ... et l'heure */
			pPtr1 = strchr(szComments, SNULL); /* Fin de la chaine */

			/* Chercher le dbut de l'heure */
			while( *pPtr1 != ' ' )
				pPtr1--;

			strcpy(szTime, pPtr1 + 1);

			/* Chercher la fin du commentaire */
			while( *pPtr1 == ' ' )
				pPtr1--;

			*++pPtr1 = SNULL;
		}

		/* Clean up de szLogger et sTime */
		pPtr1 = strchr(szTime, '\a');	//BELL
		if( pPtr1 )
			*pPtr1 = SNULL;

		//Prparer le protocole a reinjecter
		NODE_getNodeCall(0, 0, szServer);
		sprintf(pszString, "PC11^%1s^%1s^%1s^%1s^%1s^%1s^%1s^H%d^~",
			szFrequency,
			szDxCall,
			TOOLS_whatDate(szDate),
			szTime,
			szComments,
			szLogger,
			szServer,
			HOPS_fix(StreamNum, PC11, HOPS_IN, 100));
	}
#if 0
	else
	{
		pPtr1 = strstr(pszString, "***");
		if( pPtr1 && pPtr1 < (pszString + 6) )
		{
			SWITCH_discSwitch(StreamNum);
			STREAMS_disconnect(StreamNum);
		}

	}/*End IF*/
#endif

}


//-------------------------------------------------------------------------
//------ Dcoder un protocol Pavillon -------------------------------------
//-------------------------------------------------------------------------
void PROTOCOL_pavillon(int StreamNum, char * pszCmd)
{
	static char * pArg[PPARAM];
	static char   szArg[1024];
	static int    nParams;
	static int    nProtocol;
	static int    nRet;

#ifdef	DEBUGLINK_
LOG_write(LOG_BUFFERIN_FILE, LOG_DATE_NO, "[PROTOCOL_pavillon] %d> %s\n", StreamNum, pszCmd);
#endif

	/* As User ? */
	if( HOPS_fix(StreamNum, PCAS_USER, HOPS_IN, 100) )
	{
		PROTOCOL_asUser(StreamNum, pszCmd);
		if( pszCmd[2] != '1' || pszCmd[3] != '1' )
			return;
	}

	/* Ne pas deconnecter sur les chaines vides */
	if( pszCmd[0] == SNULL )
		return;

	/* Verifier que le msg est un protocole */
	if( pszCmd[0] != 'P' || pszCmd[1] != 'C' || strlen(pszCmd) > 1023 ||
	    strlen(pszCmd) < 3 )
	{
		/* Peut-tre est-ce un entte, qu'il faut alors traiter */
		if( ! strncmp(pszCmd, "<CLU-", 4) )
		{
			/* CluLink valid : lire le numro de version */
			PROTOCOL_adjacentVersion[StreamNum] = 1000 * atoi(pszCmd+5) + atoi(pszCmd+7);
			PROTOCOL_linkType[StreamNum]  = CLU_LINK;

			BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> CluLink v%d.%d implemented.\n", STREAMS_callsign[StreamNum], PROTOCOL_adjacentVersion[StreamNum]/1000, PROTOCOL_adjacentVersion[StreamNum]%1000);
			BUFFERS_addBuff(BUFFER_SCREEN1, "Transmitting local configuration...\n", OUT);

			/* Transmettre la configuration locale */
			PROTOCOL_sendInit(StreamNum);
			return;
		}

		/* L'init du link est elle OK ? */
		if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_DONE )
		{
			/* Oui -> Deconnecter. */
			SWITCH_discSwitch(StreamNum);
			STREAMS_disconnect(StreamNum);
		}
		return;
	}

	/* Extraire les arguments */
	strcpy(szArg, pszCmd);
	nParams = PROTOCOL_extractPC(szArg, pArg);

	/* Appeler la fonction correspondante */
	nProtocol = atoi(pszCmd + 2);
	nRet	  = PROTOCOL_UNKNOWN;
#ifdef	DEBUGLINK_
LOG_write(LOG_BUFFERIN_FILE, LOG_DATE_NO, "[PROTOCOL_pavillon #2] %d> nProtocol = %d\n", StreamNum, nProtocol);
#endif
	switch( nProtocol )
	{
		case PC10 :
			nRet = PAVILLON_pc10(StreamNum, nParams, pArg);
			break;

		case PC11 :
			nRet = PAVILLON_pc11(StreamNum, nParams, pArg);
			break;

		case PC12 :
			nRet = PAVILLON_pc12(StreamNum, nParams, pArg);
			break;

		case PC16 :
			nRet = PAVILLON_pc16(StreamNum, nParams, pArg);
			break;

		case PC17 :
			nRet = PAVILLON_pc17(StreamNum, nParams, pArg);
			break;

		case PC18 :
			nRet = PAVILLON_pc18(StreamNum, nParams, pArg);
			break;

		case PC19 :
			nRet = PAVILLON_pc19(StreamNum, nParams, pArg);
			break;

		case PC20 :
			nRet = PAVILLON_pc20(StreamNum, nParams, pArg);
			break;

		case PC21 :
			nRet = PAVILLON_pc21(StreamNum, nParams, pArg);
			break;

		case PC22 :
			nRet = PAVILLON_pc22(StreamNum, nParams, pArg);
			break;

		case PC23 :
			nRet = PAVILLON_pc23(StreamNum, nParams, pArg);
			break;

		case PC24 :
			nRet = PAVILLON_pc24(StreamNum, nParams, pArg);
			break;

		case PC28 :
			nRet = PAVILLON_pc28(StreamNum, nParams, pArg);
			break;

		case PC29 :
			nRet = PAVILLON_pc29(StreamNum, nParams, pArg);
			break;

		case PC30 :
			nRet = PAVILLON_pc30(StreamNum, nParams, pArg);
			break;

		case PC31 :
			nRet = PAVILLON_pc31(StreamNum, nParams, pArg);
			break;

		case PC32 :
			nRet = PAVILLON_pc32(StreamNum, nParams, pArg);
			break;

		case PC33 :
			nRet = PAVILLON_pc33(StreamNum, nParams, pArg);
			break;

		case PC34 :
			nRet = PAVILLON_pc34(StreamNum, nParams, pArg);
			break;

		case PC35 :
			nRet = PAVILLON_pc35(StreamNum, nParams, pArg);
			break;

		case PC39 :
			nRet = PAVILLON_pc39(StreamNum, nParams, pArg);
			break;

		case PC41 :
			nRet = PAVILLON_pc41(StreamNum, nParams, pArg);
			break;

		case PC44 :
			nRet = PAVILLON_pc44(StreamNum, nParams, pArg);
			break;

		case PC45 :
			nRet = PAVILLON_pc45(StreamNum, nParams, pArg);
			break;

		case PC46 :
			nRet = PAVILLON_pc46(StreamNum, nParams, pArg);
			break;

		case PC50 :
			nRet = PAVILLON_pc50(StreamNum, nParams, pArg);
			break;

		case PC51 :
			nRet = PAVILLON_pc51(StreamNum, nParams, pArg);
			break;
	}

	if( nRet != PROTOCOL_OK )
		PROTOCOL_errorPC(StreamNum, nRet, pszCmd);

}

//-------------------------------------------------------------------------
//------ Dcoder un protocol Pavillon -------------------------------------
//-------------------------------------------------------------------------
#if 0
void PROTOCOL_pavillon(int StreamNum, char * Cmd)
{
  char	sPavillon[256];
  char  sCluLink[256];
  char	sBuffer[256];
  char *pPtr1;
  char *pPtr2;

  char *pParams[PPARAM];		//PPARAM Pointeurs pour les champs cluster
  int	iNbParams;

//  char  sVersionInt[3];
  char  sNodeCall[11];
  char  sUserCall[11];

  int   iHops;
  int   index;
  int	iPos;
  unsigned short shVersion;

  char  cFlag;

  unsigned long lFrequency;

#ifdef SPECIAL_DX_
  //Ajout special permettant de se connecter a un cluster en tant qu'utili-
  //sateur pour recuperer des spots DX
  if( HOPS_fix(StreamNum, PCAS_USER, HOPS_IN, 100) )
  {
    char sLogger[20];
    char sFrequency[20];
    char sDxCall[30];
    char sComments[80];
    char sTime[20];
    char sDate[20];
    char sServer[20];

    strupr(Cmd);

    //Identifier un spot DX
    if( Cmd[0] == 'D' && Cmd[1] == 'X' && Cmd[2] == ' ')
    {
      pPtr1 = Cmd + 6;	//Pointer sur l'indicatif de l'om

      //Lire l'indicatif
      for(index = 0; index < 19; index++)
      {
	if( *pPtr1 != ':' )
	  sLogger[index] = *pPtr1++;
	else
	  break;
      }
      sLogger[index] = SNULL;
      strupr(sLogger);

      //Passer les espaces
      pPtr1++;
      while( *pPtr1 == ' ' )
	pPtr1++;

      //Lire la frquence
      for(index = 0; index < 19; index++)
      {
	if( *pPtr1 != ' ' )
	  sFrequency[index] = *pPtr1++;
	else
	  break;
      }
      sFrequency[index] = SNULL;

      //Passer les espaces
      while( *pPtr1 == ' ' )
	pPtr1++;

      //Lire l'indicatif du DX
      for(index = 0; index < 29; index++)
      {
	if( *pPtr1 != ' ' )
	  sDxCall[index] = *pPtr1++;
	else
	  break;
      }
      sDxCall[index] = SNULL;

      //Passer les espaces
      while( *pPtr1 == ' ' )
	pPtr1++;

      //Lire le commentaire
      strcpy(sComments, pPtr1);

      //y avait-il un commentaire ?
      if( strlen(sComments) < 8 )
      {
	//Non ...
	strcpy(sTime, sComments);
	strcpy(sComments, " ");
      }
      else
      {
	//Oui ... Lire le commentaire ... et l'heure
	pPtr1 = strchr(sComments, SNULL);	//Fin de la chaine

	//Chercher le dbut de l'heure
	while( *pPtr1 != ' ' )
	  pPtr1--;

	strcpy(sTime, pPtr1 + 1);

	//Chercher la fin du commentaire
	while( *pPtr1 == ' ' )
	  pPtr1--;

	*++pPtr1 = SNULL;
      }

      //Clean up de sLogger et sTime
      pPtr1 = strchr(sTime, '\a');	//BELL
      if( pPtr1 )
	*pPtr1 = SNULL;

      //Prparer le protocole a reinjecter
      NODE_getNodeCall(0, 0, sServer);
      sprintf(Cmd, "PC11^%1s^%1s^%1s^%1s^%1s^%1s^%1s^H%d^~", sFrequency,
							  sDxCall,
							  TOOLS_whatDate(sDate),
							  sTime,
							  sComments,
							  sLogger,
							  sServer,
							  HOPS_fix(StreamNum, PC11, HOPS_IN, 100));

    }
    //Rechercher des trames de deconnection
    else
    {
      pPtr1 = strstr(Cmd, "***");
      if( pPtr1 && pPtr1 < (Cmd + 6) )
      {
	SWITCH_discSwitch(StreamNum);
	STREAMS_disconnect(StreamNum);
      }
    }/*End IF*/
  }/*End IF*/
#endif

  //------ Vrifier qu'il s'agit bien d'un protocole ----------------------
  if( Cmd[0] != 'P' || Cmd[1] != 'C' || strlen(Cmd) < 2 )
  {
    //Peut-tre est-ce un entte, qu'il faut alors traiter
    if( ! strncmp(Cmd, "<CLU-", 4) )
    {
      //CluLink valid : lire le numro de version
      PROTOCOL_adjacentVersion[StreamNum] = 1000 * atoi(Cmd+5) + atoi(Cmd+7);
      PROTOCOL_linkType[StreamNum]  = CLU_LINK;

      BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> CluLink v%d.%d implemented.\n", STREAMS_callsign[StreamNum], PROTOCOL_adjacentVersion[StreamNum]/1000, PROTOCOL_adjacentVersion[StreamNum]%1000);
      BUFFERS_addBuff(BUFFER_SCREEN1, "Transmitting local configuration...\n", OUT);

      //Transmettre la configuration locale
      PROTOCOL_sendInit(StreamNum);
      return;
    }
    else if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_DONE &&
	     strlen(Cmd) > 2 )
    {
      //Le message recu n'est pas correct. Deconnecter.
      SWITCH_discSwitch(StreamNum);
      STREAMS_disconnect(StreamNum);
    }/*End IF*/
  }

  //------ Numro du protocole Pavillon -----------------------------------
  iProtNum = atoi( Cmd + 2 );

  //------ Maintenant, lire les champs ------------------------------------
  pPtr1 = Cmd + 5;		//Pointer juste aprs le numro de protocole

  for(index = 0; index < PPARAM; index++)
  {
    pParams[index] = pPtr1;

    //Rechercher le champs suivant
    while( *pPtr1 != SNULL && *pPtr1 != '^'/* && *pPtr1 != '~'*/ )
      pPtr1++;

    //Fin de la ligne de commande atteinte ?
    if( *pPtr1 == SNULL || *pPtr1 == '~' )
      break;

    *pPtr1++ = SNULL;
  }

  //Sauvegarder le nombre de parametres recus dans le protocole
  iNbParams = index;

  //Complter les champs restant par des NULLs
  for(;index < PPARAM; index++)
    pParams[index] = NULL;

  switch( iProtNum )
  {
    case PC10 :
      //Vrification de la validite du protocole
      if( iNbParams != 6 )
	return;
      if( ! TOOLS_isCall(pParams[0]) ||
	  ! TOOLS_isCall(pParams[1]) ||
	  ! TOOLS_isCall(pParams[5]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC10, HOPS_IN, 100) == 0 )
	break;
      //Lecture du flag
      if( pParams[3][0] == '*' )
	cFlag = PROTOCOL_TALK_BELL;
      else
	cFlag = 0;
      if( pParams[4][0] == ' ' )	//Link external ?
/*NON*/	PROTOCOL_sendTalk(StreamNum, pParams[5], pParams[0], "", pParams[1], pParams[2], cFlag, TOOLS_whatDateTime(), HOPS_probable(pParams[5]));
      else
/*OUI*/	PROTOCOL_sendTalk(StreamNum, pParams[5], pParams[0], pParams[1], pParams[4], pParams[2], cFlag, TOOLS_whatDateTime(), HOPS_probable(pParams[5]));
      break;

    case PC11 :
      //Vrification de la validit du protocole
      if( iNbParams != 8 )
	return;
      if( ! TOOLS_isCall(pParams[5]) || ! TOOLS_isCall(pParams[6]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC11, HOPS_IN, 100) == 0 )
	break;

      //Lire la frquence
      /*(on ajoute 0.000001 pour corriger un bug sous linux)*/
      lFrequency = (unsigned long) (atof(pParams[0]) * 10.0 + 0.000001);
      if( ! DX_checkFrequency(lFrequency) )
	return;		//Frequence incorrecte

      //Lire le hop count
      iHops = atoi( pParams[7] + 1 );
      iHops = HOPS_fix(StreamNum, PC11, HOPS_IN, iHops);

      //Enlever le "Z"  l'heure
      pParams[3][4] = SNULL;

      //Transmettre le spot aux clusters adjacents et aux utilisateurs
      PROTOCOL_sendDx(StreamNum, pParams[6], pParams[5], lFrequency, pParams[2], pParams[3], pParams[1], pParams[4], iHops);

      break;

    case PC12 :
      //Vrifier la validit du protocole
      if( iNbParams != 7 )
	return;
      if( ! TOOLS_isCall(pParams[0]) ||
	  /*! TOOLS_isCall(pParams[1]) ||*/
	  ! TOOLS_isCall(pParams[4]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC12, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = atoi( pParams[6] + 1);
      iHops = HOPS_fix(StreamNum, PC12, HOPS_IN, iHops);

      //Lecture du flag
      if( pParams[3][0] != ' ' )	//sysop flag
	cFlag = ANNOUNCE_TOSYSOP;
      else
      if( pParams[1][0] == '*' )	//cluster-wide
	cFlag = ANNOUNCE_ALL;
      else
      {
	//Est-ce pour un cluster ou une liste de distribution ?
	//ICI
	cFlag = ANNOUNCE_TOCLUSTER;	//Pour un cluster ou une liste de distribution
      }/*End IF*/

      if( pParams[5][0] == '1' )	//WX flag
	cFlag = ANNOUNCE_WX;

      //Faire suivre le protocol, et transmettre aux utilisateurs si besoin
      PROTOCOL_sendAnnounce(StreamNum, pParams[4], pParams[0], pParams[1], TOOLS_whatDateTime(), cFlag, pParams[2], iHops);

      break;

    case PC16 :
      //Verifier la validite du protocole
      if( iNbParams < 3 )	//Le nom de params est indefini, mais au moins egal a 3
	return;
      if( ! TOOLS_isCall(pParams[0]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC16, HOPS_IN, 100) == 0 )
	break;

      //Lire l'indicatif du home cluster
      strcpy(sNodeCall, pParams[0]);

      //Vrifier que ce cluster est connu
      if( ! NODE_isNodeConnected( sNodeCall ) )
      {
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : Can't add users on an unknown node (%s).\n", sNodeCall);
	break;
      }

      //Prparer les protocoles qu'il faudra envoyer aux adjacents
      sprintf(sPavillon, "PC16^%1s^", sNodeCall);
      iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(3));
      iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);

      //Lire maintenant la liste des utilisateurs
      for(index = 1; index < PPARAM-1; index++)
      {
	//Tester si la fin de la liste  t atteinte
	if( pParams[index + 1] == NULL || pParams[index + 1][0] == SNULL ) //Test sur SNULL, par scurit
	  break;

	//Lire l'indicatif
	pPtr1 = pParams[index];
	pPtr2 = sUserCall;
	while( * pPtr1 != ' ' )
	  *pPtr2++ = *pPtr1++;
	*pPtr2 = SNULL;			//Marquer la fin de la chaine

	//Lire la config utilisateur (ConfMode)
	pPtr1++;			//Passer l'espace
	cFlag = (*pPtr1++ == '*' ? PROTOCOL_USER_CONF : 0);

	//Lire la config utilisateur (Here)
	pPtr1++;			//Passer l'espace
	cFlag += (*pPtr1 == '1' ? PROTOCOL_USER_HERE : 0);

	//Ajouter cet utilisateur dans la table (si l'indicatif
	//est valide)
	if( TOOLS_isCall(sUserCall) )
	{
	  NODE_newUser(StreamNum, 0, sNodeCall, sUserCall, cFlag);

	  //Forward pour cet utilisateur ?
	  MAIL_fwd2User(StreamNum, sUserCall);

	  //Afficher le message sur la console
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Adding %s\n", sNodeCall, sUserCall);

	  //------ Ajouter au protocol Pavillon ------
	  sprintf(sBuffer, "%s %c %c^", sUserCall, (cFlag & PROTOCOL_USER_CONF ? '*' : '-'), (cFlag & PROTOCOL_USER_HERE ? '1' : '0'));
	  strcat(sPavillon, sBuffer);

	  //------ Ajouter au protocol CluLink ------
	  //Ajouter longueur indicatif et indicatif de l'utilisateur
	  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sUserCall, iCluLinkPos);
	  //Ajouter le user flag
	  sCluLink[iCluLinkPos++] = cFlag;
	}/*End IF*/
      }/*End FOR*/

      //Lecture du hop count
      iHops = atoi( pParams[index]+1 );
      iHops = HOPS_fix(StreamNum, PC16, HOPS_IN, iHops);

      //Updater le fichier des hops count (sauf si le link est en cours d'init)
      if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_DONE )
	HOPS_updateHops(pParams[0], iHops, PC16);

      //Transmettre ce protocole aux adjacents
      PROTOCOL_sendProt(StreamNum, PC16, sPavillon, sCluLink, iHops, iCluLinkPos, PROTOCOL_ALL);

      break;

    case PC17 :
      //Vrifier la validit du protocole
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) ||
	  ! TOOLS_isCall(pParams[1]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC17, HOPS_IN, 100) == 0 )
	break;

      //Effacer l'indicatif de l'utilisateur de la table
      if( ! NODE_deleteUser(StreamNum, pParams[1], pParams[0]) )
      {
	//Lecture du hop count
	iHops = atoi( pParams[2]+1 );
	iHops = HOPS_fix(StreamNum, PC17, HOPS_IN, iHops);

	//Updater le fichier des hops count
	HOPS_updateHops(pParams[1], iHops, PC17);

	//A pu tre ffac, faire suivre aux adjacents
	PROTOCOL_deleteUser(StreamNum, pParams[1], pParams[0], iHops);

	//Message sur la console
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Removing %s\n", pParams[1], pParams[0]);
      }
      else
      {
	//Erreur, cet utilisateur n'tait pas dans la table
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : Can't remove %s from node %s.\n", pParams[0], pParams[1]);
      }/*End IF*/
      break;

    case PC18 :
      //Verifier la validite du protocole
      if( iNbParams != 2 )
	return;

      /* Verifier s'il s'agit d'un init suite a la l'etablissement d'un
	 connexion ou si cela a ete demand par le sysop d l'adjacent */
      if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_DONE )
      {
	/* Demande par le sysop, informer les adjacents */
	char szReason[256];
	char szAdjacentCallSign[16];

	NODE_getNodeCall(StreamNum, 0, szAdjacentCallSign);
	sprintf(szReason, "Remote init request by %s", szAdjacentCallSign);

	PROTOCOL_clusterStreamDisconnected(StreamNum, szReason, -1);
      }

      //Dtecter si l'adjacent supporte le protocol CluLink
      pPtr1 = strstr(pParams[0], "<CLU-"); //Rechercher <CLU-x.y>

      if( pPtr1 && HOPS_fix(StreamNum, PCCLULINK, HOPS_OUT, 100) )
      {
	//CluLink valid : Lire le numro de version
//	PROTOCOL_adjacentVersion[StreamNum] = 1000 * atoi(pParams[0]+8) + atoi(pParams[0]+10);
	PROTOCOL_adjacentVersion[StreamNum] = 1000 * atoi(pPtr1+5) + atoi(pPtr1+7);

	//Accorder les versions du protocol CluLink
	if( PROTOCOL_adjacentVersion[StreamNum] > VERSION_CLU_INT )
	  PROTOCOL_adjacentVersion[StreamNum] = VERSION_CLU_INT;

	PROTOCOL_linkType[StreamNum]  = CLU_LINK;
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> CluLink v%d.%d implemented.\n", STREAMS_callsign[StreamNum], PROTOCOL_adjacentVersion[StreamNum]/1000, PROTOCOL_adjacentVersion[StreamNum]%1000);

	//Valider le protocol auprs de l'adjacent
	BUFFERS_printBuff(StreamNum, OUT, "<CLU-%d.%d>\r", PROTOCOL_adjacentVersion[StreamNum]/1000, PROTOCOL_adjacentVersion[StreamNum]%1000);
	BUFFERS_addBuff(BUFFER_SCREEN1, "Transmitting local configuration...\n", OUT);

	//Transmettre la configuration locale
	PROTOCOL_sendInit(StreamNum);
      }
      else
      {
	BUFFERS_addBuff(BUFFER_SCREEN1, "Transmitting local configuration...\n", OUT);
	PROTOCOL_adjacentVersion[StreamNum] = atoi( pParams[2] );	//Numro de version
	PROTOCOL_sendInit(StreamNum);
	BUFFERS_addBuff(StreamNum, "PC20^\n", OUT);
      }/*End IF*/
      break;

    case PC19 :
      //Verifier la validite du protocole
      if( iNbParams < 5 )	//nombre de params indetermines, mais au moins egal a 5
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC19, HOPS_IN, 100) == 0 )
	break;

      //Ne plus afficher "can't connect twice"
      SCRIPT_fileName[StreamNum][0] = SNULL;

      //Prparer les protocoles qu'il faudra forwarder aux autres adjacents
      strcpy(sPavillon, "PC19^");
      iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(1));

      for(index = 0; index < PPARAM-4; index += 4)
      {
	//Tester si la fin de la liste est atteinte
	if( pParams[index][0] == 'H' )
	  break;

	//Dtecter une boucle
	if( NODE_isNodeConnected(pParams[index + 1]) )
	{
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Loop detected : %s\n", pParams[index + 1]);
	}
	else if( TOOLS_isCall(pParams[index + 1]) )	//Verifier la validite de l'indicatif
	{
	  //Le dernier octet de NodeCall contient les params Here & ConfMode
	  cFlag  = BIT6 * (*pParams[index]   == '1' ? NODE_levelHere : NODE_levelNoHere);
	  cFlag += BIT7 * (*pParams[index+2] == '1' ? NODE_levelConf : NODE_levelNoConf);

	  //Forward pour ce cluster ?
	  MAIL_fwd(pParams[index + 1], 0L, BUFFER_SCREEN1);

	  //Ajouter le cluster dans la table
	  NODE_newNode(StreamNum, pParams[index + 1], atoi(pParams[index + 3]),  cFlag);

	  //Afficher le message sur la console
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node logged on : %s\n", pParams[index + 1]);

	  //------ Ajouter au protocole pavillon ------
	  sprintf(sBuffer, "%s^%s^%s^%s^", pParams[index], pParams[index + 1], pParams[index + 2], pParams[index + 3]);
	  strcat(sPavillon, sBuffer);

	  //------ Ajouter au protocol CluLink ------
	  //Ajouter longueur de l'indicatif et indicatif
	  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, pParams[index+1], iCluLinkPos);

	  //Ajouter Node Flag
	  sCluLink[iCluLinkPos++] = cFlag + FLAG_SOFTWAREUNKNOWN;

	  //Ajouter Version
//	  TOOLS_versionStr2Int(pParams[index+3], sVersionInt);
//	  sCluLink[iCluLinkPos++] = sVersionInt[0];
//	  sCluLink[iCluLinkPos++] = sVersionInt[1];
	  shVersion = (unsigned short) atoi(pParams[index+3]);
	  iCluLinkPos = CLULINK_addShort(sCluLink, &shVersion, iCluLinkPos);
	}/*End IF*/
      }/*End FOR*/

      //Lecture du hop count
      iHops = atoi( pParams[index]+1 );
      iHops = HOPS_fix(StreamNum, PC19, HOPS_IN, iHops);

      //Updater le fichier des hop count (sauf si le link est en cours d'init)
      //Cas particulier : updater le hops de l'adjacent lors de l'init (cela
      //ne sera plus possible plus tard)
      if( PROTOCOL_initDone[StreamNum] != PROTOCOL_INIT_DO )
      {
	if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_NOTSTART )
	  PROTOCOL_initDone[StreamNum] = PROTOCOL_INIT_DO;

	HOPS_updateHops(pParams[1], iHops, PC19);
      }

      //Transmettre ce protocole aux autres clusters
      //(seulement si au moins un des clusters n'tait pas dtect en tant que boucle)
      if( strlen( sPavillon ) > 5 )
	PROTOCOL_sendProt(StreamNum, PC19, sPavillon, sCluLink, iHops, iCluLinkPos, PROTOCOL_ALL);

      //Lancer les fwd lorsqu'une nouvelle station se connecte
      if( PROTOCOL_initDone[StreamNum] == PROTOCOL_INIT_DONE )
	MAIL_fwd();

      break;

    case PC20 :
      BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Remote initialization completed.\n", NODE_getNodeCall(StreamNum, 0, sNodeCall));
      BUFFERS_addBuff(BUFFER_SCREEN1, "Transmitting local configuration...\n", OUT);
      BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Initialization sequence completed.\n", NODE_getNodeCall(StreamNum, 0, sNodeCall));
      PROTOCOL_sendInit(StreamNum);
      BUFFERS_addBuff(StreamNum, "PC22^\n", OUT);
      PROTOCOL_initDone[StreamNum] = 2;
      MAIL_fwd();	//lancer les forward
      break;

    case PC21 :
      //Verifier la validite du protocole
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC21, HOPS_IN, 100) == 0 )
	break;

      //Enlever le cluster de la table
      if( NODE_deleteNode(StreamNum, pParams[0]) )	//Le cluster est-il bien dans la table ?
      {
	//Lecture du hop count
	iHops = atoi(pParams[2] + 1);
	iHops = HOPS_fix(StreamNum, PC21, HOPS_IN, iHops);

	//Updater le fichier des hops count
	HOPS_updateHops(pParams[0], iHops, PC21);

	//Faire suivre le protocole aux adjacents
	PROTOCOL_nodeDisconnect(StreamNum, pParams[0], pParams[1], 255, iHops);

	//Petit message sur la console
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node logged out : %s (%s)\n", pParams[0], pParams[1]);
      }
      else
      {
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : delete node for %s can't be performed.\n", pParams[0]);
      }/*End IF*/

      break;

    case PC22 :
      BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Initialization sequence completed.\n", NODE_getNodeCall(StreamNum, 0, sNodeCall));
      PROTOCOL_initDone[StreamNum] = 2;
      MAIL_fwd();	//Lancer les forward
      break;

    case PC23 :
    {
      int 	iSFI;
      int 	iA;
      int 	iK;
      char      sTime[5];

      //Verifier la validite du protocole
      if( iNbParams != 9 )
	return;
      if( ! TOOLS_isCall(pParams[6]) || ! TOOLS_isCall(pParams[7]) )
	return;
      if( strlen(pParams[1]) > 2) //Champ TIME (sans les minutes)
	return;

      //Ce protocol est-il autorise ?
      if( HOPS_fix(StreamNum, PC23, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = atoi(pParams[8] + 1);
      iHops = HOPS_fix(StreamNum, PC24, HOPS_IN, iHops);

      //Lecture de SFI, A et K
      iSFI = atoi( pParams[2] );
      iA   = atoi( pParams[3] );
      iK   = atoi( pParams[4] );

      //Lecture de l'heure (et ajouter "00" a la fin, car les minutes ne sont pas transmises)
      strcpy(sTime, pParams[1]);
      strcat(sTime, "00");

      //Faire suivre le protocole
      PROTOCOL_sendWwv(StreamNum, pParams[7]/*NodeCall*/, pParams[6]/*UserCall*/,
		       iSFI, iA, iK, pParams[0]/*Date*/, sTime,
		       pParams[5]/*Forcast*/, iHops);

    }
      break;


    case PC24 :
      //Verifier la validite de ce protocole
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC24, HOPS_IN, 100) == 0 )
	break;

      //lecture du hop count
      iHops = atoi(pParams[2] + 1);
      iHops = HOPS_fix(StreamNum, PC24, HOPS_IN, iHops);

      //Quel est l'indicatif du cluster de l'utilisateur
      if( NODE_searchUser(pParams[0], NODE_CHECKSSID_YES, sNodeCall) )	//Trouv ?
      {
	//Updater la config
	cFlag = NODE_getUserFlag(sNodeCall, pParams[0]);
	if( pParams[1][0] == '1' )
	  cFlag |= PROTOCOL_USER_HERE;
	else
	  cFlag &= ~PROTOCOL_USER_HERE;
	NODE_setUserFlag(sNodeCall, sUserCall, cFlag);

	//Transmettre l'information aux clusters adjacents
	PROTOCOL_sendUserFlag(StreamNum, sNodeCall, pParams[0], cFlag, PROTOCOL_USER_HERE, iHops);
      }/*End IF*/
      break;

    case PC28 :	//Entte d'un forward de message
    {
      char		szType[3];
      char		szBuffer[256];

      //Verifier la validite de ce protocole
      if( iNbParams != 13 )
	return;
      if( ! TOOLS_isCall(pParams[0]) ||
	  ! TOOLS_isCall(pParams[1]) ||
	  ! TOOLS_isCall(pParams[3])/* ||
	  ! TOOLS_isCall(pParams[12])*/ )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC28, HOPS_IN, 100) == 0 )
	break;

      //Liste des champs
      //pParams[0]	PC destinataire
      //pParams[1]	PC expediteur
      //pParams[2]	Recipient
      //pParams[3]	Expediteur (utilisateur)
      //pParams[4]	Date
      //pParams[5]	Heure
      //pParams[6]	Private-flag (1->prive, 0->bull)
      //pParams[7]	sujet
      //pParams[8]	?  A priori un espace
      //pParams[9]	Nombre de ligne avant un ack
      //pParams[10]	1 si Return Receipt demande, 0 sinon
      //pParams[11]	?  A priori un espace
      //pParams[12]	PC qui est a l'origine du message (StartNode)

      if( ! strcmp(pParams[12], " ") )
	pParams[12][0] = SNULL;

      //Crer le message temporaire
      //Prive ou bull ?
      if( pParams[6][0] == '1' )
	strcpy(szType, "sp");
      else
	strcpy(szType, "sb");

      //Ajouter l'indicatif de l'expediteur ...
      sprintf(szBuffer, "%s < %s", pParams[2], pParams[3]);

      if( ! MAIL_rcvRecip(StreamNum, szType, szBuffer, atoi(pParams[10])) )
	break;	//Le message n'a pas pu tre ouvert

      //Ajouter les datas concernant le mail ds le fichier temp
      if( * pParams[12] )	//Si StartNode n'est pas vide
      {
	if( ! MAIL_addSubject2Tmp(StreamNum, pParams[7], pParams[4], pParams[5],
			  pParams[12], pParams[1]) )
	break; 	//Probleme ...
      }
      else
      {
	if( ! MAIL_addSubject2Tmp(StreamNum, pParams[7], pParams[4], pParams[5],
			  "?", pParams[1]) )
	break; 	//Probleme ...
      }/*End IF*/

      //Envoyer le protocole
      PROTOCOL_sendAckSubject(StreamNum, pParams[1], pParams[0], (unsigned long) StreamNum);

      //Nombre de lignes par blocs
      PROTOCOL_mailEveryNLines[StreamNum] = (char) atoi( pParams[9] );
      PROTOCOL_mailLinesCounter[StreamNum] = PROTOCOL_mailEveryNLines[StreamNum];
    }
    break;

    case PC29 :	//Recevoir des lignes de texte (fwd)
    {
      unsigned long	lMesNum;
      char		szFileName[MAXPATH];
      FILE	       *fPtr;

      //Varifier la validite du protocole
      if( iNbParams != 4 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC29, HOPS_IN, 100) == 0 )
	break;

      //Liste des champs
      //pParams[0]      PC destinataire
      //pParams[1]	PC expediteur
      //pParams[2]	Numero du message
      //pPapams[3]	Une ligne de texte

      //Numero du message ?
      lMesNum = atol( pParams[2] );

      //Verifier validite du numero de message
      if( lMesNum < 1 || lMesNum > 64 )
	return;	//Erreur

      //Ajouter la ligne de texte ds le fichier temp
      MAIL_addText2Tmp(StreamNum, pParams[3]);

      //Gestion du conteur de ligne - faut-il envoyer un ACK ?
      if( --PROTOCOL_mailLinesCounter[StreamNum] == 0 )
      {
	PROTOCOL_sendAckText(StreamNum, pParams[1], pParams[0], lMesNum);
	PROTOCOL_mailLinesCounter[StreamNum] = PROTOCOL_mailEveryNLines[StreamNum];
      }/*End IF*/
    }
    break;

    case PC30 : //AckSubject
    case PC31 : //AckText
    //Ces deux protocoles sont grs identiquement
    {
      //Verifier le protocole
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC30, HOPS_IN, 100) == 0 )
	break;

      if( pParams[2] )	//Ne pas appeler atol si pParams[2] est null
      {
	PROTOCOL_lFwdMesNum[StreamNum] = atol (pParams[2]);
	MAIL_sendNextBloc(StreamNum, pParams[0], pParams[1], PROTOCOL_lFwdMesNum[StreamNum]);
      }
    }
    break;

    case PC32 : //CompleteText
    {
      unsigned long	lMesNum;

      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC32, HOPS_IN, 100) == 0 )
	break;

      //Numero du message ?
      lMesNum = atol( pParams[2] );

      //Verifier validite du numero de message
      if( lMesNum < 1 || lMesNum > 64 )
	return;	//Erreur

      if( MAIL_tmp2Msg( lMesNum ) )	//Valider le message
	PROTOCOL_sendAckCompleteText(StreamNum, pParams[1], pParams[0], lMesNum);

    }
    break;

    case PC33 :	//AckCompleteText
      //Verifier le protocole
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC33, HOPS_IN, 100) == 0 )
	break;

      MAIL_changeStatut(MAIL_lFwdMsgNumber[StreamNum], STATUT_F);
      MAIL_deQueue(pParams[1], MAIL_lFwdMsgNumber[StreamNum]);
      MAIL_lFwdMsgNumber[StreamNum] = 0L;
      MAIL_fwd();
      break;

    case PC41 :
      //Verifier la validite de ce protocole
      if( iNbParams != 4 )
	return;
      if( ! TOOLS_isCall(pParams[0]) )
	return;

      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC41, HOPS_IN, 100) == 0 )
	break;

      iHops = atoi(pParams[3] + 1);
      iHops = HOPS_fix(StreamNum, PC41, HOPS_IN, iHops);

      //Sauvegarder l'information
      USERS_changeConfig(pParams[0], pParams[2], atoi(pParams[1]));

      //Transmettre l'information aux clusters adjacents
      PROTOCOL_sendUserConfig(StreamNum, "", pParams[0], pParams[1], atoi(pParams[1]), iHops);
      break;

    case PC44 :
      /*Verifier la validite du protocole*/
      if( iNbParams != 6 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) ||
	  ! TOOLS_isCall(pParams[5]) )
	return;

      /*Ce protocole est-il autorise ?*/
      if( HOPS_fix(StreamNum, PC44, HOPS_IN, 100) == 0 )
	break;

      /*Repondre ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, sNodeCall);
      if( ! strcmp(sNodeCall, pParams[0]) )
      {
	/*Interroger la base de donnees*/
	if( ! DATABASE_remote(atoi(pParams[2]), pParams[3], pParams[4], pParams[1]) )
	{
	  /*Erreur acces base de donnees*/
	  PROTOCOL_dbResponse(0, pParams[0], pParams[1], atoi(pParams[2]),
			      "*** Error accessing the database : database unknown");
	}
	PROTOCOL_dbComplete(0, pParams[0], pParams[1], atoi(pParams[2]));
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbRequest(StreamNum, pParams[1], pParams[0], atoi(pParams[2]),
			   pParams[3], pParams[4], pParams[5]);
      }
      break;

    case PC45 :
      /*Verifier la validite du protocole*/
      if( iNbParams != 4 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      /*Ce protocole est t-il autorise ?*/
      if( HOPS_fix(StreamNum, PC45, HOPS_IN, 100) == 0 )
	break;

      /*Recuperer ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, sNodeCall);
      if( ! strcmp(sNodeCall, pParams[0]) )
      {
	/*Envoyer a l'utilisateur*/
	BUFFERS_printBuff(atoi(pParams[2]), OUT, "%s\n", pParams[3]);
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbResponse(StreamNum, pParams[1], pParams[0],
			    atoi(pParams[2]), pParams[3]);
      }
      break;

    case PC46 :
      /*Verifier la validite du protocole*/
      if( iNbParams != 3 )
	return;
      if( ! TOOLS_isCall(pParams[0]) || ! TOOLS_isCall(pParams[1]) )
	return;

      /*Ce protocole est t-il autorise ?*/
      if( HOPS_fix(StreamNum, PC46, HOPS_IN, 100) == 0 )
	break;

      /*Recuperer ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, sNodeCall);
      if( ! strcmp(sNodeCall, pParams[0]) )
      {
	/*Envoyer a l'utilisateur*/
	MSG_send(atoi(pParams[2]), MSG62);
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbComplete(StreamNum, pParams[1], pParams[0],
			    atoi(pParams[2]));
      }
      break;

    case PC50 :
      //Verifier la validite de ce protocole
      if( iNbParams != 3 )
	return;

      //Ce protocole est-il autorise
      if( HOPS_fix(StreamNum, PC50, HOPS_IN, 100) == 0 )
	break;

      if( ! TOOLS_isCall(pParams[0]) )
	return;

      iHops = atoi (pParams[2] + 1);
      iHops = HOPS_fix(StreamNum, PC50, HOPS_IN, iHops);

      PROTOCOL_userCount(StreamNum, pParams[0], TRUE, (unsigned short) atoi (pParams[1]), iHops);
      break;

    case PC51 :
      //Vrifier la validite de ce protocole
      if( iNbParams != 3 )
	return;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC51, HOPS_IN, 100) == 0 )
	break;

      if( ! TOOLS_isCall(pParams[0]) ||
	  ! TOOLS_isCall(pParams[1]) ||
	  ! pParams[2] )
	return;

      PROTOCOL_ping(StreamNum, pParams[1], pParams[0], (*pParams[2] == '1' ? PROTOCOL_PING_ASK : PROTOCOL_PING_ANSWER), HOPS_probable(pParams[1]) );
      break;

  }/*End SWITCH*/
}

#endif

//-------------------------------------------------------------------------
//------ Gestion des erreurs de protocole ---------------------------------
//-------------------------------------------------------------------------
void PROTOCOL_errorCL(int StreamNum, int nMessageType, int nError, char* pszCluLink)
{
	extern	char * 	PARAMS_logs;

	if( ! SET_check(PARAMS_logs, LOG_CLULINK_STR) )
		return;

	if( nError > 0 && nError < PROTOCOL_UNKNOWN )
	{
		char	szPcCall[16];

		/* Quel est le call du cluster ? */
		NODE_getNodeCall(StreamNum, 0, szPcCall);

		LOG_write(LOG_CLULINK_FILE, LOG_DATE_YES,
			  "%s (Type %d - adjacent : %s)\n%s\n---\n",
			  pszError[nError], nMessageType, szPcCall,
			  pszCluLink);
	}
}

//-------------------------------------------------------------------------
//------ Dcoder un protocol CluLink --------------------------------------
//-------------------------------------------------------------------------
void PROTOCOL_cluLink(int StreamNum, char * pszCmd)
{
	static	int	nMessageType;
	static	unsigned int nSize;
	static	char	crc;
	static	int	nRet;
	static	char*	pszCluLink;

	pszCluLink = pszCmd;

	/* Taille du protocole CluLink */
	nSize = CLULINK_getSize(pszCluLink);

	/* Passer ProtLength */
	pszCluLink += 2;

	/* Verifier le CRC */
	crc = *pszCluLink++;
	if( crc != CLULINK_fixCRC(pszCluLink, nSize) )
	{
		/* Mauvais CRC */
		 BUFFERS_printBuff(BUFFER_SCREEN1, OUT,
			"*** Protocol error : receiving of bad CRC on stream %d.\n",
			StreamNum);

		/* 3 bad CRC consecutifs. Deconnecter ! */
		if( PROTOCOL_badCRC[StreamNum]++ > 1 )
		{
			STREAMS_disconnect(StreamNum);
			SWITCH_discSwitch(StreamNum);
		}

		/* Message dans le log */
		nMessageType = (unsigned int) *pszCluLink;
		PROTOCOL_errorCL(StreamNum, nMessageType, PROTOCOL_BADCRC,
				 pszCmd);

		return;
	}
	else
	{
		/* CRC ok - reinitialiser le compteur */
		PROTOCOL_badCRC[StreamNum] = 0;
	}

	/* Identifier le type de protocole */
	nMessageType = (unsigned int) *pszCluLink++;

	/* Oter de la taille ProtLength & MessageType */
	nSize -= 2;

	/* Router le protocole */
	switch( nMessageType )
	{
		case	CL(0):
			nRet = CLULINK_cl0(StreamNum, pszCluLink, nSize);
			break;

		case	CL(1):
			nRet = CLULINK_cl1(StreamNum, pszCluLink, nSize);
			break;

		case	CL(2):
			nRet = CLULINK_cl2(StreamNum, pszCluLink, nSize);
			break;

		case	CL(3):
			nRet = CLULINK_cl3(StreamNum, pszCluLink, nSize);
			break;

		case	CL(4):
			nRet = CLULINK_cl4(StreamNum, pszCluLink, nSize);
			break;

		case	CL(5):
			nRet = CLULINK_cl5(StreamNum, pszCluLink, nSize);
			break;

		case	CL(10):
			nRet = CLULINK_cl10(StreamNum, pszCluLink, nSize);
			break;

		case	CL(11):
			nRet = CLULINK_cl11(StreamNum, pszCluLink, nSize);
			break;

		case	CL(12):
			nRet = CLULINK_cl12(StreamNum, pszCluLink, nSize);
			break;

		case	CL(13):
			nRet = CLULINK_cl13(StreamNum, pszCluLink, nSize);
			break;

		case	CL(60):
			nRet = CLULINK_cl60(StreamNum, pszCluLink, nSize);
			break;

		case	CL(64):
			nRet = CLULINK_cl64(StreamNum, pszCluLink, nSize);
			break;

		case	CL(65):
			nRet = CLULINK_cl65(StreamNum, pszCluLink, nSize);
			break;

		case	CL(66):
			nRet = CLULINK_cl66(StreamNum, pszCluLink, nSize);
			break;

		case	CL(67):
			nRet = CLULINK_cl67(StreamNum, pszCluLink, nSize);
			break;

		case	CL(95):
			nRet = CLULINK_cl95(StreamNum, pszCluLink, nSize);
			break;

		case	CL(96):
			nRet = CLULINK_cl96(StreamNum, pszCluLink, nSize);
			break;

		case	CL(97):
			nRet = CLULINK_cl97(StreamNum, pszCluLink, nSize);
			break;

		case	CL(98):
			nRet = CLULINK_cl98(StreamNum, pszCluLink, nSize);
			break;

		case	CL(99):
			nRet = CLULINK_cl99(StreamNum, pszCluLink, nSize);
			break;

		case	CL(100):
			nRet = CLULINK_cl100(StreamNum, pszCluLink, nSize);
			break;

		case	CL(160):
			nRet = CLULINK_cl160(StreamNum, pszCluLink, nSize);
			break;

		case	CL(161):
			nRet = CLULINK_cl161(StreamNum, pszCluLink, nSize);
			break;

		case	CL(162):
			nRet = CLULINK_cl162(StreamNum, pszCluLink, nSize);
			break;

		case	CL(200):
			nRet = CLULINK_cl200(StreamNum, pszCluLink, nSize);
			break;

		case	CL(201):
			nRet = CLULINK_cl201(StreamNum, pszCluLink, nSize);
			break;

		default :
			nRet = PROTOCOL_UNKNOWN;
			break;
	}

	if( nRet != PROTOCOL_OK )
		PROTOCOL_errorCL(StreamNum, nMessageType, nRet, pszCmd);
}

#if 0
//-------------------------------------------------------------------------
//------ Dcoder un protocol CluLink --------------------------------------
//-------------------------------------------------------------------------
void PROTOCOL_cluLink(int StreamNum, char * Cmd)
{
  char	sPavillon[256];
  char	sCluLink[256];
  char	sBuffer[256];
  char 	sNodeCall[11];
  char  sUserCall[11];
  char	sToUser[11];
  char	sToNode[11];
  char  sToWho[21];
  char  sReason[30];
  char  sVersionString[6];
//  char  sVersionInt[3];
  char  sDxCall[25];		//10 de plus que ncessaire, par scurit
  char  sComments[40];		//10 de plus que ncessaire, par scurit
  char	sDate[12];
  char	sTime[5];

  char 		*pPtr1;
  char 		*pEndProt;
  unsigned int	 iLength;
  unsigned short shVersion;
  unsigned short shUserCount;
  unsigned char	 	 crc;

  int 	iHops;
  int	index;
  int	iReason;
  int	iType;

  char	cFlag;

  unsigned long	lFrequency;
  unsigned long lDateTime;

  //Rerouter le protocol vers Pavillon (ce qui ne doit normallement pas
  //arriver...)
  if( Cmd[0] == 'P' && Cmd[1] == 'C' && Cmd[2] == '^' )
  {
    PROTOCOL_pavillon(StreamNum, Cmd);
    return;
  }

  //pPtr1 pointe sur le protocol CluLink
  pPtr1 = Cmd;

  //Longueur du protocol CluLink
  iLength   = CLULINK_getSize(pPtr1);
  pEndProt  = pPtr1 + iLength + 2;	//pPtr1 pointe sur la longueur du protocol
  *pEndProt = SNULL;

  //Passer ProtLength
  pPtr1 += 2;

  //Verifier le CRC
  crc = *pPtr1++;
  if( crc != CLULINK_fixCRC(pPtr1, iLength) )
  {
    //Mauvais CRC
    BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : receiving of bad CRC on stream %d.\n", StreamNum);

    //3 bad CRC consecutifs. Deconnecter !
    if( PROTOCOL_badCRC[StreamNum]++ > 1 )
    {
      STREAMS_disconnect(StreamNum);
      SWITCH_discSwitch(StreamNum);
    }

    return;
  }
  else
    PROTOCOL_badCRC[StreamNum] = 0;

  //Identifier le protocole
  switch( (unsigned char) *pPtr1++ )
  {
    case CL(0) :	//Initialization complete
      BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Remote initialization completed.\n", NODE_getNodeCall(StreamNum, 0, sNodeCall));
      PROTOCOL_initDone[StreamNum] = 2;
      MAIL_fwd();	//Lancer les forwards
      break;

    case CL(1) :	//Add Node(s)
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC19, HOPS_IN, 100) == 0 )
	break;

      //Ne plus afficher "can't connect twice"
      SCRIPT_fileName[StreamNum][0] = SNULL;

      //Prparer les protocoles qu'il faudra forwarder aux autres adjacents
      strcpy(sPavillon, "PC19^");
      iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(1));

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC19, HOPS_IN, iHops);

      //Lire maintenant la liste de nodes a ajouter
      while( pPtr1 < pEndProt )
      {
	//Lecture de l'indicatif
	pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);

	//Lecture du node flag
	cFlag = *pPtr1++;

	//Lecture du numro de version
//	sVersionInt[0] = *pPtr1++;
//	sVersionInt[1] = *pPtr1++;
//	sVersionInt[2] = SNULL;
//	TOOLS_versionInt2Str(sVersionInt, sVersionString);
	pPtr1 = CLULINK_getShort(pPtr1, &shVersion);
	sprintf(sVersionString, "%04d", shVersion);


	//Dtecter une boucle
	if( NODE_isNodeConnected( sNodeCall ) )
	{
	  //boucle dtecter - l'afficher et ignorer cet indicatif
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Loop detected : %s\n", sNodeCall);
	}
	else
	{
	  //Ajouter ce cluster dans la table
	  NODE_newNode(StreamNum, sNodeCall, atoi(sVersionString), cFlag);

	  //Updater le fichier des hops count (sauf si le link est en cours d'init)
	  //Cas particulier : updater le hops de l'adjacent, car cela ne sera
	  //plus possible plus tard ...
	  if( PROTOCOL_initDone[StreamNum] != 1 )
	  {
	    if( PROTOCOL_initDone[StreamNum] == 0 )
	      PROTOCOL_initDone[StreamNum] = 1;

	    HOPS_updateHops(sNodeCall, iHops, PC19);
	  }

	  //Forward pour ce cluster ?
	  MAIL_fwd(sNodeCall, 0L, BUFFER_SCREEN1);

	  //Afficher le message sur la console
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node logged on : %s\n", sNodeCall);

	  //------ Ajouter au protocole pavillon ------
	  sprintf(sBuffer, "%c^%s^%c^%s^", (cFlag & BIT6 ? '1' : '0'), sNodeCall, (cFlag & BIT7 ? '*' : '-'), sVersionString);
	  strcat(sPavillon, sBuffer);

	  //------ Ajouter au protocole CluLink ------
	  //Ajouter longueur de l'indicatif et indicatif
	  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);

	  //Ajouter Node Flag
	  sCluLink[iCluLinkPos++] = cFlag;

	  //Ajouter Version
//	  sCluLink[iCluLinkPos++] = sVersionInt[0];
//	  sCluLink[iCluLinkPos++] = sVersionInt[1];
	  iCluLinkPos = CLULINK_addShort(sCluLink, &shVersion, iCluLinkPos);
	}/*End IF*/
      }/*End WHILE*/

      sCluLink[iCluLinkPos] = SNULL;

      //Transmettre ce protocole aux autres autres adjacents clusters,
      //seulement si au moins un des nouveaux clusters n'a pas t
      //dtect en tant que boucle et que la longueur du protocol
      //CluLink est correcte.
      if( strlen( sPavillon ) > 5 && pPtr1 == pEndProt )
	PROTOCOL_sendProt(StreamNum, PC19, sPavillon, sCluLink, iHops, iCluLinkPos, PROTOCOL_ALL);

      break;

    case CL(2) :	//Delete Node
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC21, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC21, HOPS_IN, iHops);

      //Lecture de l'indicatif
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);

      //Updater le fichier des hops count
      HOPS_updateHops(sNodeCall, iHops, PC21);

      //Lecture de la raison
      iReason = (BYTE) *pPtr1++;

      //Raison spciale (non reconnue par CluLink)
      switch( iReason )
      {
	case 0 :
	  strcpy(sReason, PROTOCOL_REASON_STR0);
	  break;

	case 1 :
	  strcpy(sReason, PROTOCOL_REASON_STR1);
	  break;

	case 2 :
	  strcpy(sReason, PROTOCOL_REASON_STR2);
	  break;

	case 255 : //Raison particulire,  lire
	  pPtr1 = PROTOCOL_readPascalString(sReason, pPtr1);
	  break;

	default : //Sais pas laquelle
	  iReason = 255;
	  strcpy(sReason, "Unknown");
	  break;
      }/*End SWITCH*/

      //Vrifier que la longueur du protocol reue est conforme
      if( pPtr1 == pEndProt )
      {
	//Enlever le cluster de la table
	if( NODE_deleteNode(StreamNum, sNodeCall) )	//Le cluster est-il bien dans la table ?
	{
	  //Faire suivre le protocol aux adjacents
	  PROTOCOL_nodeDisconnect(StreamNum, sNodeCall, sReason, iReason, iHops);

	  //Petit message sur la console
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node logged out : %s (%s)\n", sNodeCall, sReason);
	}
	else
	{
	  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : delete node for %s can't be performed.\n", sNodeCall);
	}/*End IF*/
      }/*End IF*/

      break;

    case CL(3) :	//Add user(s)
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC16, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC16, HOPS_IN, iHops);

      //Lecture de l'indicatif du home node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);

      //Updater le fichier des hop count (sauf si le link est en cours d'init)
      if( PROTOCOL_initDone[StreamNum] == 2 )
	HOPS_updateHops(sNodeCall, iHops, PC16);

      //Vrifier que ce cluster est connu
      if( ! NODE_isNodeConnected( sNodeCall ) )
      {
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : Can't add users on an unknown node (%s).\n", sNodeCall);
	break;
      }
      //Prparer les protocoles qu'il faudra envoyer aux adjacents
      sprintf(sPavillon, "PC16^%s^", sNodeCall);
      iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(3));
      iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);
      //Lire maintenant la liste des utilisateurs  ajouter
      while( pPtr1 < pEndProt )
      {
	//Lecture de l'indicatif
	pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
	//Lecture du user flag
	cFlag = *pPtr1++;
	//Forward pour cet utilisateur ?
	MAIL_fwd2User(StreamNum, sUserCall);
	//Ajouter cet utilisateur dans la table
	NODE_newUser(StreamNum, 0, sNodeCall, sUserCall, cFlag);
	//Afficher le message sur la console
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Adding %s\n", sNodeCall, sUserCall);
	//------ Ajouter au protocol Pavillon ------
	sprintf(sBuffer, "%s %c %c^", sUserCall, (cFlag & PROTOCOL_USER_CONF ? '*' : '-'), (cFlag & PROTOCOL_USER_HERE ? '1' : '0'));
	strcat(sPavillon, sBuffer);
	//------ Ajouter au protocol CluLink ------
	//Ajouter longueur indicatif et indicatif de l'utilisateur
	iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sUserCall, iCluLinkPos);
	//Ajouter le user flag
	sCluLink[iCluLinkPos++] = cFlag;
      };/*End WHILE*/
      //Transmettre ce protocole aux adjacents
      sCluLink[iCluLinkPos] = SNULL;
      PROTOCOL_sendProt(StreamNum, PC16, sPavillon, sCluLink, iHops, iCluLinkPos, PROTOCOL_ALL);
      break;

    case CL(4) :	//Delete user
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC17, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC17, HOPS_IN, iHops);

      //Lecture de l'indicatif du home node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);

      //Updater le fichier des hops count
      HOPS_updateHops(sNodeCall, iHops, PC17);

      //Lecture de l'indicatif de l'utilisateur
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Effacer l'indicatif de l'utilisateur de la table
      if( ! NODE_deleteUser(StreamNum, sNodeCall, sUserCall) )
      {
	//A pu tre ffac, faire suivre aux adjacents
	PROTOCOL_deleteUser(StreamNum, sNodeCall, sUserCall, iHops);
	//Message sur la console
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "Node %s> Removing %s\n", sNodeCall, sUserCall);
      }
      else
      {
	//Erreur, cet utilisateur n'tait pas dans la table
	BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Protocol error : Can't remove %s from node %s.\n", sUserCall, sNodeCall);
      }/*End IF*/
      break;

    case CL(10) :	//Change user flag
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC24, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC24, HOPS_IN, iHops);
      //Lecture de l'indicatif du home node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'utilisateur
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture du user flag

      //Updater et regarder quelle info a change
      cFlag = NODE_getUserFlag(sNodeCall, sUserCall);
      if( (cFlag & PROTOCOL_USER_HERE) != (*pPtr1 & PROTOCOL_USER_HERE) )
	iType = PROTOCOL_USER_HERE;
      else if( (cFlag & PROTOCOL_USER_CONF) != (*pPtr1 & PROTOCOL_USER_CONF) )
	iType = PROTOCOL_USER_CONF;
      cFlag = *pPtr1++;
      NODE_setUserFlag(sNodeCall, sUserCall, cFlag);

      //Transmettre l'information aux clusters adjacents
      if( iType )
	PROTOCOL_sendUserFlag(StreamNum, sNodeCall, sUserCall, cFlag, iType, iHops);
      break;

    case CL(11) :	//Change user data
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC41, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC41, HOPS_IN, iHops);

      //Lecture de l'indicatif du node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'utilisateur
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture du type d'information (data type)
      iType = (unsigned int) *pPtr1++;
      //Lecture de la chaine d'information
      pPtr1 = PROTOCOL_readPascalString(sBuffer, pPtr1);
      //Sauvegarder l'information
      USERS_changeConfig(sUserCall, sBuffer, iType);
      //Transmettre l'information aux clusters adjacents
      PROTOCOL_sendUserConfig(StreamNum, sNodeCall, sUserCall, sBuffer, iType, iHops);
      break;

    case CL(12) :	//Protocol CluLink
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC51, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC51, HOPS_IN, iHops);

      //Lecture de l'indicatif du node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'utilisateur
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      //Lecture du ping flag
      cFlag = *pPtr1++;

      //Faire suivre le protocole ... ou repondre si besoin.
      PROTOCOL_ping(StreamNum, sNodeCall, sToNode, (cFlag == 1 ? PROTOCOL_PING_ASK : PROTOCOL_PING_ANSWER), iHops);
      break;

    case CL(13) :	//User count
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC50, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC50, HOPS_IN, iHops);

      //Lecture de l'indicatif de l'expditeur
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture du user count
      pPtr1 = CLULINK_getShort(pPtr1, &shUserCount);

      //Faire suivre le protocole ...
      PROTOCOL_userCount(StreamNum, sNodeCall, TRUE, shUserCount, iHops);
      break;

    case CL(64) :    	//Dx spot
      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC11, HOPS_IN, iHops);

      //Lecture de l'indicatif du home node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif du logger
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture de la frquence
      pPtr1 = CLULINK_getLong(pPtr1, &lFrequency);
      //Lecture de l'indicatif du DX
      pPtr1 = PROTOCOL_readPascalString(sDxCall, pPtr1);
      //Lecture de la date et de l'heure
      pPtr1 = CLULINK_getLong(pPtr1, &lDateTime);
      //Lecture du commentaire
      pPtr1 = PROTOCOL_readPascalString(sComments, pPtr1);
      //Convertir lDateTime en chaines
      TOOLS_dateTime_Long2Str(sDate, sTime, lDateTime);
      //Faire suivre le protocole aux autres adjacents et utilisateurs
      PROTOCOL_sendDx(StreamNum, sNodeCall, sUserCall, lFrequency, sDate, sTime, sDxCall, sComments, iHops);
      break;

    case CL(65) :	//Spot WWV
    {
      unsigned long	lDatas;
      unsigned int	iSfi,
			iA,
			iK;
      char		sForecast[256];  //Pour une MAX de 80

      //Ce protocol est-il autorise ?
      if( HOPS_fix(StreamNum, PC23, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC23, HOPS_IN, iHops);

      //Lecture de l'indicatif du home node
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif du logger
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture de SFI, A et K
      pPtr1 = CLULINK_getLong(pPtr1, &lDatas);
      iSfi = lDatas 	    & 511;      //9 premiers bits
      iA   = (lDatas >> 9)  & 511; 	//9 bits suivants
      iK   = (lDatas >> 18) & 15;	//4 derniers bits
      //Lecture de la date et de l'heure
      pPtr1 = CLULINK_getLong(pPtr1, &lDateTime);
      //Lecture de la prevision
      pPtr1 = PROTOCOL_readPascalString(sForecast, pPtr1);
      //Convertir lDateTime en chaines
      TOOLS_dateTime_Long2Str(sDate, sTime, lDateTime);
      //Faire suivre le protocole aux autres adjacents et utilisateurs
      PROTOCOL_sendWwv(StreamNum, sNodeCall, sUserCall, iSfi, iA, iK, sDate, sTime, sForecast, iHops);
    }
      break;

    case CL(66) :     	//Annonce
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC12, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;
      iHops = HOPS_fix(StreamNum, PC12, HOPS_IN, iHops);

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'utilisateur qui a expdi cette annonce
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Ajouter l'indicatif du node destinataire ou de la liste de distribution SI NECESSAIRE (ANNOUNCE_ALL est a 0)
      pPtr1 = PROTOCOL_readPascalString(sToWho, pPtr1);
      //Lecture de la date et de l'heure
      pPtr1 = CLULINK_getLong(pPtr1, &lDateTime);
      //Lecture du Flag
      cFlag = (unsigned int) *pPtr1++;
      //Lecture du message
      pPtr1 = PROTOCOL_readPascalString(sBuffer, pPtr1);
      //Transmettre et afficher l'annonce
      PROTOCOL_sendAnnounce(StreamNum, sNodeCall, sUserCall, sToWho, lDateTime, cFlag, sBuffer, iHops);
      break;

    case CL(67) :	//Talk
      //Ce protocol est-il autoris ?
      if( HOPS_fix(StreamNum, PC10, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture de l'indicatif du cluster destinataire
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      //Lecture de l'indicatif du destinataire
      pPtr1 = PROTOCOL_readPascalString(sToUser, pPtr1);
      //Lecture de la date
      pPtr1 = CLULINK_getLong(pPtr1, &lDateTime);
      //Lecture du flag
      cFlag = (unsigned int) *pPtr1++;
      //Lecture du message
      pPtr1 = PROTOCOL_readPascalString(sBuffer, pPtr1);
      //Faire suivre le talk (ventuellement l'afficher)
      PROTOCOL_sendTalk(StreamNum, sNodeCall, sUserCall, sToNode, sToUser, sBuffer, cFlag, lDateTime, iHops);
      break;

    case CL(95) :	//SendSubject
    {
      char		sStartNode[10];
      char		sTitle[61];
      char		szType[3];
      char		cNumLines;
      BYTE		byRetReceipt;

      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC28, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur du msg
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      //Lecture de l'indicatif du cluster destinataire (MOI)
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      //Lecture de l'indicatif du recipient
      pPtr1 = PROTOCOL_readPascalString(sToUser, pPtr1);
      //Lecture du StartNode du msg
      pPtr1 = PROTOCOL_readPascalString(sStartNode, pPtr1);
      //Lecture du titre du msg
      pPtr1 = PROTOCOL_readPascalString(sTitle, pPtr1);
      //Lecture de la date
      pPtr1 = CLULINK_getLong(pPtr1, &lDateTime);
      //Lecture du flag
      cFlag = (unsigned int) *pPtr1++;

      cNumLines = cFlag & 31;	//Bits 0-4

      if( cFlag & 64 )	//Tester le bit 6 (return receipt)
	byRetReceipt = TRUE;
      else
	byRetReceipt = FALSE;

      if( cFlag & 128 )	//Tester le bit 7 (perso)
	strcpy(szType, "sp");
      else
	strcpy(szType, "sb");

      //Convertir lDateTime en chaines
      TOOLS_dateTime_Long2Str(sDate, sTime, lDateTime);

      //Creer le fichier temporaire
      if( ! MAIL_rcvRecip(StreamNum, szType, sToUser, byRetReceipt) )
	break;	//Le message n'a pas pu tre ouvert

      //Ajouter les datas concernant le mail ds le fichier temp
      if( ! MAIL_addSubject2Tmp(StreamNum, sTitle, sDate, sTime, sStartNode,
				sNodeCall) )
	break;

      //Envoyer le protocole
      PROTOCOL_sendAckSubject(StreamNum, sNodeCall, sToNode, (unsigned long) StreamNum);

      //Nombre de lignes par blocs
      PROTOCOL_mailEveryNLines[StreamNum] = cNumLines;
      PROTOCOL_mailLinesCounter[StreamNum] = PROTOCOL_mailEveryNLines[StreamNum];
    }
    break;

    case CL(96) :	//AckSubject
    case CL(98) :	//AckText
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC30, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur du msg
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);

      PROTOCOL_lFwdMesNum[StreamNum] = (unsigned long) StreamNum; //Pour y mettre quelque chose
      MAIL_sendNextBloc(StreamNum, sToNode, sNodeCall, PROTOCOL_lFwdMesNum[StreamNum]);

      break;

    case CL(97) :	//SendText
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC30, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur du msg
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      //une ligne de texte
      pPtr1 = PROTOCOL_readPascalString(sBuffer, pPtr1);

      //Ajouter la ligne de texte ds le fichier temp
      MAIL_addText2Tmp(StreamNum, sBuffer);

      //Gestion du conteur de ligne - faut-il envoyer un ACK ?
      if( --PROTOCOL_mailLinesCounter[StreamNum] == 0 )
      {
	PROTOCOL_sendAckText(StreamNum, sNodeCall, sToNode, (unsigned long) StreamNum);
	PROTOCOL_mailLinesCounter[StreamNum] = PROTOCOL_mailEveryNLines[StreamNum];
      }/*End IF*/

      break;

    case CL(99) :	//CompleteText
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC32, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur du msg
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);

      if( MAIL_tmp2Msg( StreamNum ) )	//Valider le message
	PROTOCOL_sendAckCompleteText(StreamNum, sNodeCall, sToNode, (unsigned long) StreamNum);

      break;

    case CL(100) :	//AckCompleteText
      //Ce protocole est-il autorise ?
      if( HOPS_fix(StreamNum, PC33, HOPS_IN, 100) == 0 )
	break;

      //Lecture du hop count
      iHops = (unsigned int) *pPtr1++;

      //Lecture de l'indicatif du cluster d'origine
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      //Lecture de l'indicatif de l'expditeur du msg
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);

      MAIL_changeStatut(MAIL_lFwdMsgNumber[StreamNum], STATUT_F);
      MAIL_deQueue(sNodeCall, MAIL_lFwdMsgNumber[StreamNum]);
      MAIL_lFwdMsgNumber[StreamNum] = 0L;
      MAIL_fwd();
      break;

    case CL(160) :	/*Remote DB request PC44*/
    {
      int      nUserStream;
      char     szQualifier[256];
      char     szKey[256];
      char     szServer[16];

      /*Ce protocole est t-il autorise ?*/
      if( HOPS_fix(StreamNum, PC44, HOPS_IN, 100) == 0 )
	break;

      /*Lecture du hop count*/
      iHops = (unsigned int) *pPtr1++;

      /*Cluster d'origine*/
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      /*Cluster destinataire*/
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      /*Utilisateur*/
      pPtr1 = PROTOCOL_readPascalString(sUserCall, pPtr1);
      /*Base de donnees - "Qualifier"*/
      pPtr1 = PROTOCOL_readPascalString(szQualifier, pPtr1);
      /*Clef - "Key"*/
      pPtr1 = PROTOCOL_readPascalString(szKey, pPtr1);
      /*Numero du stream de l'utilisateur*/
      nUserStream = (unsigned int) *pPtr1++;

      /*Repondre ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, szServer);
      if( ! strcmp(sToNode, szServer) )
      {
	/*Interroger la base de donnees*/
	if( ! DATABASE_remote(nUserStream, szQualifier, szKey, sNodeCall) )
	{
	  /*Erreur acces database*/
	  PROTOCOL_dbResponse(0, sToNode, sNodeCall, nUserStream,
		"*** Error accessing the database : database unknown");
	}
	PROTOCOL_dbComplete(StreamNum, sToNode, sNodeCall, nUserStream);
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbRequest(StreamNum, sNodeCall, sToNode, nUserStream,
			   szQualifier, szKey, sUserCall);
      }
    }
    break;

    case CL(161) :	/*Remote DB response PC45*/
    {
      int      nUserStream;
      char     szString[256];
      char     szServer[16];

      /*Ce protocole est t-il autorise ?*/
      if( HOPS_fix(StreamNum, PC45, HOPS_IN, 100) == 0 )
	break;

      /*Lecture du hop count*/
      iHops = (unsigned int) *pPtr1++;

      /*Cluster d'origine*/
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      /*Cluster destinataire*/
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      /*Chaine contenant la reponse*/
      pPtr1 = PROTOCOL_readPascalString(szString, pPtr1);
      /*Numero du stream de l'utilisateur*/
      nUserStream = (unsigned int) *pPtr1++;

      /*Repondre ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, szServer);
      if( ! strcmp(sToNode, szServer) )
      {
	/*Transmettre a l'utilisateur*/
	BUFFERS_printBuff(nUserStream, OUT, "%s\n", szString);
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbResponse(StreamNum, sNodeCall, sToNode, nUserStream,
			    szString);
      }
    }
    break;

    case CL(162) :	/*Remote DB complete PC46*/
    {
      int      nUserStream;
      char     szServer[16];

      /*Ce protocole est t-il autorise ?*/
      if( HOPS_fix(StreamNum, PC45, HOPS_IN, 100) == 0 )
	break;

      /*Lecture du hop count*/
      iHops = (unsigned int) *pPtr1++;

      /*Cluster d'origine*/
      pPtr1 = PROTOCOL_readPascalString(sNodeCall, pPtr1);
      /*Cluster destinataire*/
      pPtr1 = PROTOCOL_readPascalString(sToNode, pPtr1);
      /*Numero du stream de l'utilisateur*/
      nUserStream = (unsigned int) *pPtr1++;

      /*Repondre ou transmettre a un adjacent*/
      NODE_getNodeCall(0, 0, szServer);
      if( ! strcmp(sToNode, szServer) )
      {
	/*Transmettre a l'utilisateur*/
	MSG_send(nUserStream, MSG62);
      }
      else
      {
	/*Faire suivre ...*/
	PROTOCOL_dbComplete(StreamNum, sNodeCall, sToNode, nUserStream);
      }
    }
    break;

  }/*End SWITCH*/
}

#endif

/*-----------------------------------------------------------------------
  ------ Transmettre l'initialisation (PC19 et PC16) --------------------
  -----------------------------------------------------------------------*/
void PROTOCOL_sendInit(int StreamNum)
{
	if( PROTOCOL_nFixedInit )
	{
		/* Initialisation du link corrigee */
		CONINIT_sendInitProtocols(StreamNum);
	}
	else
	{
		/* Initialisation du link buggee (idem a Pavillon) */
		char  	sCluLink[256];
		char	sNodeCall[11];
		char  	sUserCall[11];
		char  	sVersionString[10];
		unsigned short shVersion;
		char  	cFlag;

		int   	iNodeStream = 0;
		int	iNodePos = 0;
		int   	iUserPos = 0;
		int	iCount = 0;
		int 	iHops;
	
		/* Transmettre les Clusters */

		/* Parcourir les streams, un  un */
		for(iNodeStream = 0; iNodeStream <= 64; iNodeStream++)
		{
			/* Est-ce un cluster ? (ATTENTION, ne pas rebalancer la config de l'autre !) */
			if( iNodeStream != StreamNum && (iNodeStream == 0 || STREAMS_level[iNodeStream] & LEVEL_cluster) )
			{
				/* Parcourir toutes les positions pour ce stream */
				for(iNodePos = 0; iNodePos < 100; iNodePos++)
				{
					NODE_getNodeCall(iNodeStream, iNodePos, sNodeCall);

					/* Transmettre le protocole PC19 */
					if( sNodeCall[0] )
					{
						/* Definir le hop count */
						iHops = HOPS_fix(StreamNum, PC19, HOPS_OUT, 100);

						/* L'indicatif de MON cluster doit OBLIGATOIREMENT tre transmis
						   mme en cas de "link external" */
						if(iNodeStream == 0 && iNodePos == 0 && iHops == 0)
							iHops = HOPS_fix(StreamNum, PCEXTERNAL, HOPS_OUT, 100);

						/* Lecture du numero de version et du Flag */
						cFlag = NODE_getNodeFlag(iNodeStream, iNodePos);
						NODE_getNodeVersion(iNodeStream, iNodePos, sVersionString);

						if( PROTOCOL_linkType[StreamNum] == PAVILLON )
						{
							/* Creer le protocole pavillon */
							if( iCount == 0 )
								BUFFERS_addBuff(StreamNum, "PC19^", OUT);

							BUFFERS_printBuff(StreamNum, OUT, "%c^%s^%c^%s^", (cFlag & BIT6 ? '1' : '0'), sNodeCall, (cFlag & BIT7 ? '1' : '0'), sVersionString);

							if( HOPS_fix(StreamNum, PCEXTERNAL, HOPS_OUT, 100) )
							{
								/* C'est un link external, ne transmettre que MON cluster */
								BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", iHops);
								return;
							}

							if( iCount++ == 12 )
							{
								BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", iHops);
								iCount = 0;
							}/*End IF*/
						}
						else
						{
							/* Creer le protocole clulink */
							if( iCount == 0 )
							{
								/* Definir le protocol et le hop count */
								iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(1));
								sCluLink[4] = (BYTE) iHops;
							}

							/* Ajouter longueur indicatif et indicatif */
							iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);
							/* Ajouter le node flag */
							sCluLink[iCluLinkPos++] = cFlag;
							/* Ajouter numro de version */
							shVersion = (unsigned short) atoi(sVersionString);
							iCluLinkPos = CLULINK_addShort(sCluLink, &shVersion, iCluLinkPos);

							/* Transmettre par blocs et 12 et gerrer le PB des links external */
							if( iCount++ == 12 || HOPS_fix(StreamNum, PCEXTERNAL, HOPS_OUT, 100) )
							{
								/* Longueur du protocole */
								CLULINK_putSize(sCluLink, iCluLinkPos - 2);
								sCluLink[0] |= CL_BROADCAST;	/* Mettre a 1 le bit BROADCAST */
								BUFFERS_addBuff_sized(StreamNum, sCluLink, OUT, iCluLinkPos);
								iCount = 0;

								if( HOPS_fix(StreamNum, PCEXTERNAL, HOPS_OUT, 100) )
									return;	/* Ne transmettre que MON cluster (ce qui est fait), donc arrter ICI */
							}
						}/*End IF*/
					}
					else
						break;
				}/*End FOR*/
			}/*End IF*/
		}/*End FOR*/

		if( iCount != 0 )
		{
			if( PROTOCOL_linkType[StreamNum] == PAVILLON )
				BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", HOPS_fix(StreamNum, PC19, HOPS_OUT, 100) );
			else
			{
				/* Longueur du protocole */
				CLULINK_putSize(sCluLink, iCluLinkPos - 2);
				sCluLink[0] |= CL_BROADCAST;	/* Mettre a 1 le bit BROADCAST */
				BUFFERS_addBuff_sized(StreamNum, sCluLink, OUT, iCluLinkPos);
			}/*End IF*/
		}/*End IF*/

		/* Transmettre les Utilisateurs */
		/* Definir le hop count */
		iHops = HOPS_fix(StreamNum, PC16, HOPS_OUT, 100);

		/* Reinitialiser le compteur */
		iCount = 0;

		/* Inutile d'aller plus loin si le hop count est null */
		if( iHops == 0 )
			return;

		/* Parcourir les streams, un  un */
		for(iNodeStream = 0; iNodeStream <= 64; iNodeStream++)
		{
			/* Est-ce un cluster ? (ATTENTION, ne pas rebalancer la config de l'autre !) */
			if( iNodeStream != StreamNum && (iNodeStream == 0 || STREAMS_level[iNodeStream] & LEVEL_cluster) )
			{
				/* Parcourir toutes les positions pour ce stream */
				for(iNodePos = 0; iNodePos < 100; iNodePos++)
				{
					NODE_getNodeCall(iNodeStream, iNodePos, sNodeCall);

					/* Transmettre le protocole PC16 */
					if( sNodeCall[0] )
					{
						/* Parcourir la table pour trouver les indicatifs */
						for(iUserPos = 1; iUserPos <= 64; iUserPos++)
						{
							/* Lire l'indicatif (s'il existe) */
							NODE_getUserCall(iNodeStream, iNodePos, iUserPos, sUserCall);

							/* Y a t-il un utilisateur connecte ? */
							if( sUserCall[0] != SNULL )
							{
								/* Lire la config de cet utilisateur */
								cFlag = NODE_getUserFlag(iNodeStream, iNodePos, iUserPos);

								/* Selon le protocol mis en place avec l'adjacent */
								if( PROTOCOL_linkType[StreamNum] == PAVILLON )
								{
									/* Creer le protocol PAVILLON */
									if( iCount == 0 )
										BUFFERS_printBuff(StreamNum, OUT, "PC16^%s^", sNodeCall);

									BUFFERS_printBuff(StreamNum, OUT, "%s %c %c^", sUserCall, (cFlag & PROTOCOL_USER_CONF ? '*' : '-'), (cFlag & PROTOCOL_USER_HERE ? '1' : '0') );

									if( iCount++ == 16 )
									{
										BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", iHops);
										iCount = 0;
									}/*End IF*/
								}
								else
								{
									/* Creer le protocol CluLink */
									if( iCount == 0 )
									{
										/* Definir le protocol et le hop count */
										iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(3));
										sCluLink[4] = (BYTE) iHops;

										/* Ajouter longueur indicatif et indicatif du node */
										iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);
									}/*End IF*/

									/* Ajouter longueur indicatif et indicatif de l'utilisateur */
									iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sUserCall, iCluLinkPos);

									/* Ajouter le user flag */
									sCluLink[iCluLinkPos++] = cFlag;

									if( iCount++ == 16 )
									{
										/* Longueur du protocole */
										CLULINK_putSize(sCluLink, iCluLinkPos - 2);
										sCluLink[0] |= CL_BROADCAST;	/* Mettre a 1 le bit BROADCAST */
										BUFFERS_addBuff_sized(StreamNum, sCluLink, OUT, iCluLinkPos);
										iCount = 0;
									}/*End IF*/
								}/*End IF*/
							}/*End IF*/
						}/*End FOR*/
					}/*End IF*/

					/* Transmettre le protocol pour ce node (selon le protocol mis en place avec l'adjacent) */
					if( iCount != 0 )	/* Sauf s'il n'y a pas d'utilisateurs */
					{
						if( PROTOCOL_linkType[StreamNum] == PAVILLON )
							BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", iHops);
						else
						{
							CLULINK_putSize(sCluLink, iCluLinkPos - 2);
							sCluLink[0] |= CL_BROADCAST;	/* Mettre a 1 le bit BROADCAST */
							BUFFERS_addBuff_sized(StreamNum, sCluLink, OUT, iCluLinkPos);
						}
						iCount = 0;
					}
				}/*End FOR*/
			}/*End IF*/
		}/*End FOR*/

		if( iCount != 0 )
		{
			if( PROTOCOL_linkType[StreamNum] == PAVILLON )
				BUFFERS_printBuff(StreamNum, OUT, "H%d^\n", iHops);
			else
			{
				/* Longueur du protocole */
				CLULINK_putSize(sCluLink, iCluLinkPos - 2);
				sCluLink[0] |= CL_BROADCAST;	//Mettre a 1 le bit BROADCAST
				BUFFERS_addBuff_sized(StreamNum, sCluLink, OUT, iCluLinkPos);
			}/*End IF*/
		}/*End IF*/
	} /* End IF */

	/* Protocol CluLink : Indiquer que l'initialisation est termine */
	if( PROTOCOL_linkType[StreamNum] == CLU_LINK )
	{
		char 	szCluLink[256];
		int     nCluLinkPos;

		nCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(0));
		CLULINK_putSize(szCluLink, nCluLinkPos - 2);
		BUFFERS_addBuff_sized(StreamNum, szCluLink, OUT, nCluLinkPos);
	}
}

//---------------------------------------------------------------------------
//------ Crer l'entte d'un protocole CluLink ------------------------------
//---------------------------------------------------------------------------
unsigned int PROTOCOL_setCluLink(char * CluLink, int ProtNum)
{
  CluLink[0] = 255;	//CluLink size
  CluLink[1] = 255;	//CluLink size
  CluLink[2] = 255;	//CRC
  CluLink[3] = ProtNum;	//#Protocole
  CluLink[4] = 99; 	//Hop count
  CluLink[5] = SNULL;

  return 5;		//Position du pointeur iCluLinkPos
}

//---------------------------------------------------------------------------
//------ Informer les adjacents qu'un node a t dconnect -----------------
//---------------------------------------------------------------------------
void PROTOCOL_nodeDisconnect(int StreamNum, char * NodeCall, char * sReason, int iReason, int Hops)
{
  char 	sPavillon[256];
  char	sCluLink[256];
 
  //Limiter la raison  40 caractres
/*  if( strlen(sReason) > 29 )	//Faire ce test, car sReason peut provenir de * pParams
    sReason[29] = SNULL;*/
    TOOLS_maxLength(sReason, 40);

  //------ Protocol Pavillon ------
  sprintf(sPavillon, "PC21^%s^%1s^", NodeCall, sReason);

  //------ Protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(2));

  //Ajouter longueur indicatif et indicatif
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, NodeCall, iCluLinkPos);

  //Raison
  sCluLink[iCluLinkPos++] = (BYTE) iReason;

  //Si la raison n'est pas prdfinie (en particulier si le protocole
  //provient d'un link pavillon ...
  if( iReason == 255 )
  {
    //Ajouter longueur de la raison, et la raison
    iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sReason, iCluLinkPos);
  }/*End IF*/

  PROTOCOL_sendProt(StreamNum, PC21, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, -1);
}

//---------------------------------------------------------------------------
//------ Ajouter une chaine du type Pascal  une chaine de caractres -------
//---------------------------------------------------------------------------
//La fonction retourne la nouvelle longueur de Dest
int PROTOCOL_addPascalString(char * Dest, char * Org, unsigned int Pos)
{
  char * pPtr = Org;

  //Longueur de la chaine
  Dest[Pos++] = (BYTE) strlen( Org );

  //Ajouter la chaine
  while( *pPtr != SNULL )
    Dest[Pos++] = *pPtr++;

  return Pos;
}

//--------------------------------------------------------------------------
//------ Extraire une chaine du type Pascal --------------------------------
//--------------------------------------------------------------------------
//La fonction retourne un pointeur pointant sur la caractre qui suit la
//chaine Pascal
char * PROTOCOL_readPascalString(char * Dest, char * Org, int nMaxSize)
{
	int	index;
	int 	nSize;

	nSize = *Org++;

	if( nMaxSize != 0 && nSize > nMaxSize )
		return NULL;

	for(index = 0; index < nSize; index++)
		Dest[index] = *Org++;
	Dest[index] = SNULL;

	return Org;
}

char * PROTOCOL_readPascalString(char * Dest, char * Org)
{
	return PROTOCOL_readPascalString(Dest, Org, 0);
}

//--------------------------------------------------------------------------
//------ Informer les adjacents de la connection d'un utilisateur local ----
//--------------------------------------------------------------------------
void PROTOCOL_sendNewLocalUser(char * UserCall, unsigned char Flag)
{
  char sPavillon[256];
  char sCluLink[256];
  char sNodeCall[11];

  //Lire l'indicatif de MON cluster
  NODE_getNodeCall(0, 0, sNodeCall);

  //------ Crer le protocol Pavillon ------
  sprintf(sPavillon, "PC16^%s^%s - %c^", sNodeCall, UserCall, (Flag & PROTOCOL_USER_HERE ? '1' : '0'));

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(3));
  //Ajouter mon indicatif
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);
  //Ajouter l'indicatif de l'utilisateur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, UserCall, iCluLinkPos);
  //Ajouter le cFlag
  sCluLink[iCluLinkPos++] = Flag;

  //Envoyer les protocols aux adjacents
  PROTOCOL_sendProt(0, PC16, sPavillon, sCluLink, 100, iCluLinkPos, PROTOCOL_ALL, -1);
}

//--------------------------------------------------------------------------
//------ Informer les adjacents de la dconnection d'un utilisateur --------
//--------------------------------------------------------------------------
void PROTOCOL_deleteUser(int StreamNum, char * NodeCall, char * UserCall, int Hops)
{
  char 	sPavillon[256];
  char	sCluLink[256];
  //------ Crer le protocol Pavillon ------
  sprintf(sPavillon, "PC17^%s^%s^", UserCall, NodeCall);

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(4));
  //Ajouter mon indicatif
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, NodeCall, iCluLinkPos);
  //Ajouter l'indicatif de l'utilisateur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, UserCall, iCluLinkPos);

  //Envoyer les protocols aux adjacents
  PROTOCOL_sendProt(StreamNum, PC16, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, -1);
}

//--------------------------------------------------------------------------
//------ Informer les adjacents de la dconnection d'un utilisateur local --
//--------------------------------------------------------------------------
void PROTOCOL_deleteLocalUser(char * UserCall)
{
  char sNodeCall[11];

  //Lire l'indicatif de MON cluster
  NODE_getNodeCall(0, 0, sNodeCall);

  PROTOCOL_deleteUser(0, sNodeCall, UserCall, 100);
}

//--------------------------------------------------------------------------
//------ Envoyer un DX -----------------------------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendDx(int StreamNum, char * NodeCall, char * UserCall, unsigned long Frequency, char * Date, char * Time, char * DxCall, char * Comments, int Hops)
{
  unsigned long lDateTime;
  char		sPavillon[256];
  char		sCluLink[256];
  char		sBuffer[256];
  char		sLogger[11];
  extern int	PARAMS_iSetDxSsid;
  int		nWAZ;

  //PacketCluster ne veut pas de champ vide
  if( *Comments == SNULL )
    strcpy(Comments, " ");

  /* Filtre entrant */
  nWAZ = FILTER_searchWAZ(NodeCall);
  if( ! FILTER_isOK(StreamNum, FILTER_IN, nWAZ) )
	return;		/* Filtre */

  //Verifier que le spot n'est pas deja logge ...
  if( DX_isDupe(DxCall, Frequency, Comments, Time, UserCall) )
    return;

  //Enregistrer le DX
  DX_add(DxCall, Frequency, Comments, Time, Date, UserCall, NodeCall);

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(64));

  //Ajouter l'indicatif du cluster
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, NodeCall, iCluLinkPos);

  //Ajouter l'indicatif de l'utilisateur (logger)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, UserCall, iCluLinkPos);

  //Ajouter la frquence
  iCluLinkPos = CLULINK_addLong(sCluLink, &Frequency, iCluLinkPos);

  //Ajouter l'indicatif du DX
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, DxCall, iCluLinkPos);

  //Ajouter la date et l'heure
//  lDateTime = TOOLS_whatDateTime();
  lDateTime = TOOLS_dateTime_Str2Long(Date, Time);
  iCluLinkPos = CLULINK_addLong(sCluLink, &lDateTime, iCluLinkPos);

  //Ajouter le commentaire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Comments, iCluLinkPos);

  //------ Crer le protocol Pavillon ------
  sprintf(sPavillon, "PC11^%ld.%ld^%1s^%s^%sZ^%1s^%s^%s^", Frequency/10L,
							 Frequency%10L,
							 DxCall,
							 Date,
							 Time,
							 Comments,
							 UserCall,
							 NodeCall);


  //Envoyer le protocol
  PROTOCOL_sendProt(StreamNum, PC11, sPavillon, sCluLink, Hops, iCluLinkPos,
                    PROTOCOL_ALL, nWAZ);

  //Maintenant gnrer l'information  transmettre aux utilisateurs
  strcpy(sLogger, UserCall);
  strcat(sLogger, ":");

  /* Faut il enlever le SSID du logger ? */
  if( PARAMS_iSetDxSsid == FALSE )
	TOOLS_removeSsid(sLogger);

  sprintf(sBuffer, "DX de %-7s %10.1f  %-12s %-30s %4sZ%c%c\n", sLogger,
								(float) Frequency / 10,
								DxCall,
								Comments,
								Time,
								BELL,
								BELL);
  USERS_sendAll(sBuffer, USERS_ALL, nWAZ);
}

//---------------------------------------------------------------------------
//------ Transmettre une configuration utilisateur --------------------------
//---------------------------------------------------------------------------
void PROTOCOL_sendUserConfig(int StreamNum, char * FromNode, char * FromUser, char * String, int Data, int Hops)
{
  char	sPavillon[256];
  char	sCluLink[256];
  char	sFromNode[11];

  //------ Crer le protocole PAVILLON ------
  if( Data == USERS_LOCATOR )
  {
	LOCATOR_PAVCOORDINATES pavCoordinates;
	LOCATOR_locator2pavCoordinates(String, &pavCoordinates);
	sprintf(sPavillon, "PC41^%s^3^%d %d %c %d %d %c^",
        		   FromUser,
			   pavCoordinates.nLatDegrees,
			   pavCoordinates.nLatMinutes,
			  (pavCoordinates.nLatN ? 'N' : 'S'),
			   pavCoordinates.nLongDegrees,
			   pavCoordinates.nLongMinutes,
			  (pavCoordinates.nLongE ? 'E' : 'W'));
  }
  else
  	sprintf(sPavillon, "PC41^%s^%d^%1s^", FromUser, Data, String);

  //Rechercher l'indicatif du home node (s'il n'est pas prcis lors de l'appel de la fonction)
  if( *FromNode == SNULL || *FromNode == '?' )
  {
    //Rechercher dans la table
    if( ! NODE_searchUser(FromUser, NODE_CHECKSSID_YES, sFromNode) )
    {
      //L'indicatif n'a pas t trouv
      strcpy(sFromNode, "?");
    }

    //Ceci ne peut normallement pas arriver, mais on teste quand meme ...
    if( *sFromNode == SNULL )
      strcpy(sFromNode, "?");
  }
  else
    strcpy(sFromNode, FromNode);

  /* Si c'est une info lat/long, transmettre le locator en lieu et place */
  if( Data == USERS_COORDINATES )
  {
	char szLocator[8];

	/* Convertir */
	if( ! LOCATOR_pavStr2locator(String, szLocator) )
		return;	/* Probleme sur le nbr d'arg */

	strcpy(String, szLocator);

	Data = USERS_LOCATOR;
  }

  //------ Crer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(11));
  //Ajouter l'indicatif du cluster expediteur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sFromNode, iCluLinkPos);
  //Ajouter l'indicatif de l'utilisateur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromUser, iCluLinkPos);
  //Ajouter data type (type d'information)
  sCluLink[iCluLinkPos++] = (BYTE) Data;
  //Ajouter la chaine d'information
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, String, iCluLinkPos);

  //Envoyer les protocols aux adjacents
  PROTOCOL_sendProt(StreamNum, PC41, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre une annonce -------------------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendAnnounce(int StreamNum, char * FromNode, char * FromUser, char * ToNode, unsigned long DateTime, char Flag, char * String, int Hops)
{
  char	sToWho[21];
  char	sPavillon[256];
  char	sCluLink[256];
  char	sBuffer[256];
  int	iStreamOfNode;
  int	iNodePos;
  int	nWAZ;

  /* Filtre entrant */
  nWAZ = FILTER_searchWAZ(FromNode);
  if( ! FILTER_isOK(StreamNum, FILTER_IN, nWAZ) )
	return;		/* Filtre */

  //Limiter la longueur du message  80 caractres
  TOOLS_maxLength(String, 200);

  /* Mettre a jour la variable sToWho */
  if( Flag & ANNOUNCE_ALL )
	strcpy(sToWho, "ALL");
  else if( Flag & ANNOUNCE_LOCAL )
	strcpy(sToWho, "LOCAL");
  else if( Flag & ANNOUNCE_TOSYSOP )
	strcpy(sToWho, "SYSOP");
  else if( Flag & ANNOUNCE_TOCLUSTER )
	strcpy(sToWho, "CLUSTER");

  /* Verifier que l'annonce n'est pas deja loggee */
  if( ANNOUNCE_isDupe(FromNode, sToWho, FromUser, String, DateTime) )
	return;

  //------ Crer le protocole PAVILLON ------
  sprintf(sPavillon, "PC12^%s^%s^%1s^%c^%s^%c^", FromUser, ToNode, String, (Flag & ANNOUNCE_TOSYSOP ? '*' : ' '), FromNode, (Flag & ANNOUNCE_WX ? '1' : '0'));

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(66));
  //Ajouter l'indicatif du cluster expditeur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromNode, iCluLinkPos);
  //Ajouter l'indicatif de l'utilisateur expditeur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromUser, iCluLinkPos);
  //Ajouter l'indicatif du node destinataire ou de la liste de distribution SI NECESSAIRE (ANNOUNCE_ALL est a 0)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);
  //Ajouter la date
  iCluLinkPos = CLULINK_addLong(sCluLink, &DateTime, iCluLinkPos);
  //Ajouter le flag
  sCluLink[iCluLinkPos++] = (BYTE) Flag;
  //Ajouter le message
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, String, iCluLinkPos);

  /* Preparer le message */
  sprintf(sBuffer, "%s de %s: %s%c\n", sToWho, FromUser, String, BELL);

  /* A qui faut-il adresser le message ? */
  if( Flag & ANNOUNCE_ALL || Flag & ANNOUNCE_LOCAL )
  {
	/* Afficher le message aux utilisateurs */
	USERS_sendAll(sBuffer, USERS_ALL, nWAZ);
  }
  else if( Flag & ANNOUNCE_TOSYSOP )
  {
	/* Au sysops seulement */
	USERS_sendAll(sBuffer, USERS_SYSOP, nWAZ);
  }
  else if( Flag & ANNOUNCE_TOCLUSTER )
  {
	/* Est ce pour ici ? */
	if( ! strcmp(ToNode, NODE_getNodeCall(0, 0, sBuffer)) )
	{
		/* OUI : adresser le message aux utilisateurs connectes */
		USERS_sendAll(sBuffer, USERS_ALL, nWAZ);
	}
	else
	{
		/* Sur quel stream le cluster destinataire est-il connecte ? */
		if( NODE_isNodeConnected(ToNode, &iStreamOfNode, &iNodePos) )
			PROTOCOL_sendProt(StreamNum, PC12, sPavillon, sCluLink, Hops, iCluLinkPos, iStreamOfNode, nWAZ);
	}/* End IF */
  }/* End IF */

  /* Envoyer les protocoles aux adjacents (sauf si c'est une annonce locale, ou adressee a un cluster en particulier) */
  if( ! (Flag & ANNOUNCE_LOCAL) && ! (Flag & ANNOUNCE_TOCLUSTER) )
	PROTOCOL_sendProt(StreamNum, PC12, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, nWAZ);

  /* Enregistrer l'annonce */
  ANNOUNCE_add(FromNode, sToWho, FromUser, Flag, String, DateTime);
}

#if 0 /* Reecriture de la fonction TALK : 20 feb 2000 */
//---------------------------------------------------------------------------
//------ Transmettre un talk ------------------------------------------------
//---------------------------------------------------------------------------
int PROTOCOL_sendTalk(int StreamNum, char * FromNode, char * FromUser, char * ToNode, char * ToUser, char * String, char Flag, unsigned long DateTime, int Hops)
{
  char 	sNodeCall[11];
  int	iIsExt = FALSE;
  char	sPavillon[256];
  char	sCluLink[256];
  int   iStream;
  int	iNodePos;
  char	sTime[10];

  //Limiter la longueur du message  80 caractres
  TOOLS_maxLength(String, 200);

  //sNodeCall sera utilise pour designer le cluster destinataire
  strcpy(sNodeCall, ToNode);

  //Rechercher l'utilisateur dans la liste
  if( strchr(ToUser, '-') )	//Un SSID est-il precise ?
    iStream = NODE_searchUser(ToUser, NODE_CHECKSSID_YES, sNodeCall);
  else
    iStream = NODE_searchUser(ToUser, NODE_CHECKSSID_NO, sNodeCall);

  //Si le SSID 0 a ete precise par l'utilisateur, le virrer
  if( TOOLS_whatSsid(ToUser) == 0 )
    TOOLS_removeSsid(ToUser);

  //De meme, si le SSID de l'utilisateur est 0
  if( TOOLS_whatSsid(FromUser) == 0 )
    TOOLS_removeSsid(FromUser);

  //L'utilisateur a t-il t trouv ?
  if( iStream )
  {
    //Est-ce un utilisateur local ?
    if( ! *sNodeCall )
    {
      //Oui, adresser le talk a cet utilisateur
	  if( USERS_getFlags(iStream) & USERS_NOBEEP )/*BEEP ou NOBEEP ?*/
	BUFFERS_printBuff(iStream, OUT, "TALK de %s <%sZ>: %s\n", FromUser, TOOLS_whatTime(sTime), String);
	  else
		BUFFERS_printBuff(iStream, OUT, "TALK de %s <%sZ>: %s%c\n", FromUser, TOOLS_whatTime(sTime), String, BELL);

      return TRUE;
    }/*End IF*/
  }
  else
  {
    //Non, et ToNode est-il prcis ?
    if( ! *ToNode )
      return FALSE;		//Non, donc on ne sait pas ou diriger le talk

    if( ! NODE_isNodeConnected(ToNode, &iStream, &iNodePos) )
      return FALSE;

    iIsExt = TRUE;		//Oui, mais dans ce cas, protocol external
    strcpy(sNodeCall, ToNode);
  }/*End IF*/

  //------ Crer le protocole PAVILLON ------
  if( iIsExt )
      sprintf(sPavillon, "PC10^%s^%s^%1s^%c^%s^%s^", FromUser, sNodeCall, String, (Flag & PROTOCOL_TALK_BELL ? '*' : ' '), ToUser, FromNode);
  else
      sprintf(sPavillon, "PC10^%s^%s^%1s^%c^ ^%s^",  FromUser, ToUser,    String, (Flag & PROTOCOL_TALK_BELL ? '*' : ' '),         FromNode);

  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(67));

  //Ajouter l'indicatif du cluster de dpart
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromNode, iCluLinkPos);
  //Ajouter l'indicatif de l'auteur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromUser, iCluLinkPos);
  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sNodeCall, iCluLinkPos);
  //Ajouter l'indicatif du destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToUser, iCluLinkPos);
  //Ajouter la date et l'heure
  iCluLinkPos = CLULINK_addLong(sCluLink, &DateTime, iCluLinkPos);
  //Ajouter le flag
  sCluLink[iCluLinkPos++] = Flag;
  //Ajouter le message
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, String, iCluLinkPos);

  //Transmettre le protocole
  PROTOCOL_sendProt(StreamNum, PC10, sPavillon, sCluLink, Hops, iCluLinkPos, iStream);

  return TRUE;
}
#endif

/*-------------------------------------------------------------------------
  ------ Transmettre un protocole talk ------------------------------------
  -------------------------------------------------------------------------*/
void PROTOCOL_sendTalkProt(int StreamNum, char * szFromNode, char * szFromUser, char * szToNode, char * szToUser, char * szString, char Flag, unsigned long lDateTime, int nExternal, int nHops)
{
	char szPavillon[1024];
	char szCluLink[1024];
	int  nCluLinkPos;

	/* Creer le protocole PAVILLON */
	if( nExternal )
	{
		sprintf(szPavillon, "PC10^%s^%s^%1s^%c^%s^%s^",
			szFromUser,
			szToNode,
			szString,
			(Flag & PROTOCOL_TALK_BELL ? '*' : ' '),
			szToUser,
			szFromNode);
	}
	else
	{
		sprintf(szPavillon, "PC10^%s^%s^%1s^%c^ ^%s^",
			szFromUser,
			szToUser,
			szString, (
			Flag & PROTOCOL_TALK_BELL ? '*' : ' '),
			szFromNode);
	}

	/* Protocol CluLink */
	nCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(67));

	/* Indicatif du cluster de depart */
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromNode, nCluLinkPos);
	/* Indicatif de l'auteur */
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromUser, nCluLinkPos);
	/* Indicatif du cluster destinataire */
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szToNode,   nCluLinkPos);
	/* Indicatif du destinataire */
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szToUser,   nCluLinkPos);
	/* Date et heure */
	nCluLinkPos = CLULINK_addLong(szCluLink, &lDateTime, nCluLinkPos);
	/* Flag */
	szCluLink[nCluLinkPos++] = Flag;
	/* Message */
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szString, nCluLinkPos);

	/* Transmettre le protocole */
	PROTOCOL_sendProt(0, PC10, szPavillon, szCluLink, nHops, nCluLinkPos, StreamNum, -1);
}

/*-------------------------------------------------------------------------
  ------ Transmettre un talk ----------------------------------------------
  -------------------------------------------------------------------------*/
int PROTOCOL_sendTalk(int StreamNum, char * FromNode, char * FromUser, char * ToNode, char * ToUser, char * String, char Flag, unsigned long DateTime, int Hops)
{
	char szToNode[16];		/* Cluster destinataire */
	char szToUser[16];		/* Utilisateur destinataire */
	char szToUser_nossid[16];       /* Idem mais sans ssid */
	int  nStream;
	int  nNodePos;
	int  nUserPos;
	int  nDone = FALSE;

	/* Cluster dest. */
	TOOLS_maxLength(ToNode, 9);
	strcpy(szToNode, ToNode);

	/* Utilisateur dest. */
	TOOLS_maxLength(ToUser, 9);
	strcpy(szToUser, ToUser);
	strcpy(szToUser_nossid, ToUser);
	TOOLS_removeSsid(szToUser_nossid);

	/* Limiter la longueur du message a 200 caracteres */
	TOOLS_maxLength(String, 200);

	/* Rechercher l'utilisateur destinataire - balayer toutes les
	   tables */
	for(nStream = 0; nStream <= LastStream; nStream++)
	{
		for(nNodePos = 0; nNodePos < NODE_MAXANODE; nNodePos++)
		{
			char szNodeCall[16];

			/* Indicatif node */
			NODE_getNodeCall(nStream, nNodePos, szNodeCall);
			if( szNodeCall[0] == SNULL )
				break;	/* fin de table atteinte */

			/* Parcourir la liste des utilisateurs */
			for(nUserPos = 0; nUserPos <= 65; nUserPos++)
			{
				char szUserCall[16];

				NODE_getUserCall(nStream, nNodePos, nUserPos, szUserCall);
				if( szUserCall[0] == SNULL )
					continue; /* Pas d'utilisateur ici */

				if( nStream == 0 ) /* Utilisateur local ? */
				{
					char cBellFlag = ' ';
					int  nUserStream;
					char szTime[16];

					TOOLS_removeSsid(szUserCall);

					if( strcmp(szUserCall, szToUser_nossid) )
						continue;

					/* Stream de l'utilisateur */
					nUserStream = nUserPos;// - ANODE_USEROFFSET;

					/* Bingo : adresser le message au destinataire */
					if( USERS_getFlags(nUserStream) & USERS_NOBEEP )
						cBellFlag = BELL;

					/* Adresser le talk */
					BUFFERS_printBuff(nUserStream, OUT, "TALK de %s <%sZ>: %s%c\n",
							  FromUser,
							  TOOLS_whatTime(szTime),
							  String,
							  cBellFlag);
					nDone = TRUE;
				}
				else	/* Recherche de l'utilisateur sur le reseau */
				{
					if( strcmp(szUserCall, szToUser) )
						continue;

					/* Router le protocole */
					if( nStream == StreamNum )
						continue;	/* loop */
					PROTOCOL_sendTalkProt(nStream, FromNode, FromUser, szNodeCall,
							      szToUser, String, Flag, DateTime, FALSE, Hops);

					nDone = TRUE;

				} /* End IF */
			} /*End FOR nUserPos */
		} /* End FOR nNodePos */
	} /* End FOR nStream */

	if( nDone == FALSE && szToNode[0] != SNULL )
	{
		/* External - verifier que l'adjacent existe */
		if( ! NODE_isNodeConnected(szToNode, &nStream, &nNodePos) )
			return FALSE;

		PROTOCOL_sendTalkProt(nStream, FromNode, FromUser, szToNode,
				      szToUser, String, Flag, DateTime, TRUE, Hops);
		return TRUE;
	}

	return nDone;
}

//---------------------------------------------------------------------------
//------ Transmettre une config utilisateur ---------------------------------
//---------------------------------------------------------------------------
void PROTOCOL_sendUserFlag(int StreamNum, char * NodeCall, char * UserCall, char Flag, int Type, int Hops)
{
  char	sPavillon[256];
  char	sCluLink[256];
  int	iPc = -1;

  //------ Crer le protocol Pavillon ------
  if( Type == PROTOCOL_USER_HERE )
  {
    sprintf(sPavillon, "PC24^%s^%c^", UserCall, (Flag & PROTOCOL_USER_HERE ? '1' : '0'));
    iPc = PC24;
  }
  else if( Type == PROTOCOL_USER_CONF )
  {
    if( Flag & PROTOCOL_USER_CONF )
    {
      sprintf(sPavillon, "PC13^%s^", UserCall);
      iPc = PC13;
    }
    else
    {
      sprintf(sPavillon, "PC14^%s^", UserCall);
      iPc = PC14;
    }/*End IF*/
  }/*End IF*/

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(10));
  //Ajouter l'indicatif du node
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, NodeCall, iCluLinkPos);
  //Ajouter l'indicatif de l'utilisateur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, UserCall, iCluLinkPos);
  //Ajouter le user flag
  sCluLink[iCluLinkPos++] = Flag;

  //Envoyer les protocoles aux adjacents
  PROTOCOL_sendProt(StreamNum, iPc, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, -1);
}


//---------------------------------------------------------------------------
//------ Gestion des PINGS --------------------------------------------------
//---------------------------------------------------------------------------
//La fonction retourne TRUE si le cluster destinataire a ete trouve et FALSE sinon
int PROTOCOL_ping(int StreamNum, char * FromNode, char * ToNode, int PingFlag, int Hops)
{
  char	sMyNodeCall[11];
  char  sPavillon[256];
  char	sCluLink[256];
  int	iStream;
  int	iNodePos;

  //L'indicatif du serveur
  NODE_getNodeCall(0, 0, sMyNodeCall);

  //Le ping est-il pour moi ?
  if( ! strcmp(ToNode, sMyNodeCall) )
  {

    //Quelle direction ?
    if( PingFlag == PROTOCOL_PING_ASK )
    {
      //Renvoyer le ping dans sa direction

      //------ Creer le protocole Pavillon ------
      sprintf(sPavillon, "PC51^%s^%s^0^", FromNode, ToNode);

      //------ Creer le protocole CluLink ------
      iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(12));
      //Ajouter l'indicatif du cluster de depart (cad moi)
      iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);
      //Ajouter l'indicatif du cluster destinataire
      iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromNode, iCluLinkPos);
      //Ajouter le PingFlag
      sCluLink[iCluLinkPos++] = 0;
      //Transmettre le protocole
      PROTOCOL_sendProt(0, PC51, sPavillon, sCluLink, Hops, iCluLinkPos, StreamNum, -1);
    }
    else
    {
      //Envoyer le calcul du temps aller-retour a l'utilisateur
      PING_sendUser(FromNode);
    }

    return TRUE;
  }
  else
  {
    //Le PING n'etait pas pour moi. Faire suivre ...

    //Le cluster destinataire est-il connecte ?
    if( ! NODE_isNodeConnected(ToNode, &iStream, &iNodePos) )
      return FALSE;

    //------ Creer le protocole Pavillon ------
    sprintf(sPavillon, "PC51^%s^%s^%c^", ToNode, FromNode, (PingFlag == PROTOCOL_PING_ASK ? '1' : '0') );

    //------ Creer le protocole CluLink ------
    iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(12));
    //Ajouter l'indicatif du cluster de depart
    iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FromNode, iCluLinkPos);
    //Ajouter l'indicatif du cluster destinataire
    iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);
    //Ajouter le PingFlag
    sCluLink[iCluLinkPos++] = (unsigned char) PingFlag;

    //Transmettre le protocole
    PROTOCOL_sendProt(StreamNum, PC51, sPavillon, sCluLink, Hops, iCluLinkPos, iStream, -1);

    return TRUE;
  }/*End IF*/
}

//--------------------------------------------------------------------------
//------ Envoyer un WWV ----------------------------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendWwv(int StreamNum, char * NodeCall, char * UserCall, int SFI, int A, int K, char * Date, char * Time, char * Forecast, int Hops)
{
  unsigned long lDateTime;
  unsigned long lDatas;
  char		sPavillon[256];
  char		sCluLink[256];
  char		sBuffer[256];
  char		sLogger[11];
  char		szTime[8];
  int		nWAZ;

  //PacketCluster ne veut pas de champ vide
  if( *Forecast == SNULL )
    strcpy(Forecast, " ");

  nWAZ = FILTER_searchWAZ(NodeCall);
  if( ! FILTER_isOK(StreamNum, FILTER_IN, nWAZ) )
	return;		/* Filtre */

  if( ! WWV_isAllowed(Date, Time) )
	return;

  //Enregistrer le DX
  WWV_add(NodeCall, UserCall, Forecast, Date, Time, SFI, A, K);

  //------ Crer le protocol CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(65));

  //Ajouter l'indicatif du cluster
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, NodeCall, iCluLinkPos);

  //Ajouter l'indicatif de l'utilisateur (logger)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, UserCall, iCluLinkPos);

  //Ajouter SFI, K et A
  //Bits 0-8   : SFI (9 bits -> 512)
  //Bits 9-17  : A   (9 bits -> 512)
  //Bits 18-21 : K   (4 bits -> 16)
  lDatas  = ((unsigned long) K) << 18;
  lDatas += ((unsigned long) A) << 9;
  lDatas +=  (unsigned long) SFI;
  iCluLinkPos = CLULINK_addLong(sCluLink, &lDatas, iCluLinkPos);

  //Ajouter la date et l'heure
  lDateTime = TOOLS_dateTime_Str2Long(Date, Time);
  iCluLinkPos = CLULINK_addLong(sCluLink, &lDateTime, iCluLinkPos);

  //Ajouter la prevision (forecast)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Forecast, iCluLinkPos);

  //------ Crer le protocol Pavillon ------
  if( Time[0] == '0' )
  {
	szTime[0] = Time[1];
	szTime[1] = SNULL;
  }
  else
  {
	szTime[0] = Time[0];
	szTime[1] = Time[1];
	szTime[2] = SNULL;
  }
  sprintf(sPavillon, "PC23^%s^%s^%d^%d^%d^%1s^%s^%s^", Date,
						       szTime,
						       SFI,
						       A,
						       K,
						       Forecast,
						       UserCall,
						       NodeCall);
  //Envoyer le protocol
  PROTOCOL_sendProt(StreamNum, PC23, sPavillon, sCluLink, Hops, iCluLinkPos, PROTOCOL_ALL, nWAZ);

  //Maintenant gnrer l'information  transmettre aux utilisateurs
  strcpy(sLogger, UserCall);
  strcat(sLogger, ":");

  sprintf(sBuffer, "WWV de %-7s SFI=%d A=%d K=%d %-30s %4sZ%c%c\n", sLogger,
								    SFI,
								    A,
								    K,
								    Forecast,
								    Time,
								    BELL,
								    BELL);
  USERS_sendAll(sBuffer, USERS_ALL, nWAZ);
}

//--------------------------------------------------------------------------
//------ Transmettre un SendSubject ----------------------------------------
//--------------------------------------------------------------------------
//Streamnum indique LE NUMERO DU STREAM sur lequel doit etre forwarde le MAIL
void PROTOCOL_sendSubject(int StreamNum, char * ToNode, char * FmNode, char * Recipient, char * Sender, char * Date, char * Time, int Private, char * Subject, int NumOfLines, char * StartNode, BYTE RetReceipt)
{
  unsigned long	lDateTime;
  char  sPavillon[256];
  char	sCluLink[256];
  unsigned char	cFlag;

  //Initialiser
  PROTOCOL_lFwdMesNum[StreamNum] = 0L;

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC28^%1s^%1s^%1s^%1s^%1s^%sZ^%c^%1s^ ^%d^%d^ ^%1s^",
		     ToNode,
		     FmNode,
		     Recipient,
		     Sender,
		     Date,
		     Time,
		     (Private == TYPE_P ? '1' : '0'),
		     Subject,
		     NumOfLines,
		     RetReceipt,
		     StartNode);


  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(95));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif de l'expediteur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Sender, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Ajouter l'indicatif du recipient
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Recipient, iCluLinkPos);

  //Ajouter l'indicatif du StartNode
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, StartNode, iCluLinkPos);

  //Ajouter le titre
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Subject, iCluLinkPos);

  //Ajouter la date et l'heure
  lDateTime = TOOLS_dateTime_Str2Long(Date, Time);
  iCluLinkPos = CLULINK_addLong(sCluLink, &lDateTime, iCluLinkPos);

  //Ajouter le Message Flag
  if( NumOfLines > 31 )	//Nombre de bits limites a 4 (0-4)
    NumOfLines = 31;
  cFlag  = (unsigned char) NumOfLines;
  cFlag |= (Private == TYPE_P ? 128 : 0);	//perso
  cFlag |= (RetReceipt == TRUE ? 64 : 0);	//return receipt
  sCluLink[iCluLinkPos++] = cFlag;

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC28, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}


//--------------------------------------------------------------------------
//------ Transmettre l'Ack-Subject (mail fwd) ------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendAckSubject(int StreamNum, char * ToNode, char * FmNode, unsigned long MesNum)
{
  char  sPavillon[256];
  char	sCluLink[256];

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC30^%s^%s^%ld^", ToNode, FmNode, MesNum);

#ifdef	DEBUGLINK_
LOG_write(LOG_BUFFERIN_FILE, LOG_DATE_NO, "[PROTOCOL_sendAckSubject #1] %d> Acces fonction - protocole : %s\n", StreamNum, sPavillon);
#endif


  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(96));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC30, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre du texte ----------------------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendText(int StreamNum, char * ToNode, char * FmNode, unsigned long MesNum, char * Text)
{
  char  sPavillon[256];
  char	sCluLink[256];

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC29^%1s^%1s^%ld^%1s^", ToNode, FmNode, MesNum, Text);

  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(97));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Ajouter le texte
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, Text, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC29, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}


//--------------------------------------------------------------------------
//------ Transmettre l'Ack-Text (mail fwd) ---------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendAckText(int StreamNum, char * ToNode, char * FmNode, unsigned long MesNum)
{
  char  sPavillon[256];
  char	sCluLink[256];

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC31^%s^%s^%ld^", ToNode, FmNode, MesNum);

  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(98));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC31, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre CompleteText ------------------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendCompleteText(int StreamNum, char * ToNode, char * FmNode, unsigned long MesNum)
{
  char  sPavillon[256];
  char	sCluLink[256];

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC32^%s^%s^%ld^", ToNode, FmNode, MesNum);

  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(99));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC32, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}


//--------------------------------------------------------------------------
//------ Transmettre l'Ack-CompleteText (mail fwd) -------------------------
//--------------------------------------------------------------------------
void PROTOCOL_sendAckCompleteText(int StreamNum, char * ToNode, char * FmNode, unsigned long MesNum)
{
  char  sPavillon[256];
  char	sCluLink[256];

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC33^%s^%s^%ld^", ToNode, FmNode, MesNum);

  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(100));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, FmNode, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, ToNode, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC33, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre un Abort Forwarding -----------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_abortForwarding(int StreamNum)
{
  char	sToNode[10];
  char	sMyNodeCall[10];
  char  sPavillon[256];
  char	sCluLink[256];

  //Rechercher l'indicatif de l'adjacent
  NODE_getNodeCall(StreamNum, 0, sToNode);
  NODE_getNodeCall(0,         0, sMyNodeCall);

  //Message sur la console
  BUFFERS_printBuff(BUFFER_SCREEN1, OUT, "*** Forwarding timeout with %s.", sToNode);

  //Ne pas envoyer de timeout si le numero de message est encore inconnu
  if( ! PROTOCOL_lFwdMesNum[StreamNum] )
    return;

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC42^%s^%s^%ld^", sToNode, sMyNodeCall, PROTOCOL_lFwdMesNum[StreamNum]);

  //------ Creer le protocole CluLink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(101));

  //Ajouter l'indicatif du Cluster (le mien)
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sMyNodeCall, iCluLinkPos);

  //Ajouter l'indicatif du cluster destinataire
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sToNode, iCluLinkPos);

  //Envoyer le protocol
  PROTOCOL_sendProt(0, PC42, sPavillon, sCluLink, 100, iCluLinkPos, StreamNum, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre le user count (PC50) ----------------------------------
//--------------------------------------------------------------------------
//iToAll indique si le protocole doit etre adresse a TOUS les adjacents (cas
//d'un protocole recu d'un autre cluster) ou seulement a un seul
//(cas ou c'est CE cluster qui envoie le PC50)
void PROTOCOL_userCount(int StreamNum, char * sFromNode, int iToAll, unsigned short shUserCount, int iHops)
{
  char sPavillon[256];
  char sCluLink[256];

  if( iToAll == FALSE )	//Requete locale, calculer le nombre de users a envoyer
  {
    int	iNodes,
	iLocalUsers,
	iTotalUsers,
	iMaxUsers;

    NODE_getConfig(iNodes, iLocalUsers, iTotalUsers, iMaxUsers);

    if( HOPS_isExternal(StreamNum) )
    {
      char szAdjacentCall[16];

      /* Link external, transmettre le nombre d'utilisateur total du
	 reseau moins (bien sur) le nombre d'utilisateurs connectes
	 au link adjacent */
      NODE_getNodeCall(StreamNum, 0, szAdjacentCall);
      iTotalUsers -= NODE_getUserCount(StreamNum, szAdjacentCall);

      shUserCount = (unsigned short) iTotalUsers;
    }
    else
    {
      //Link normal, transmettre seulement les localusers
      shUserCount = (unsigned short) iLocalUsers;
    }
  }
  else
  {
    NODE_setUserCount(StreamNum, sFromNode, (int) shUserCount);
  }

  //------ Creer le protocole Pavillon ------
  sprintf(sPavillon, "PC50^%s^%d^", sFromNode, shUserCount);

  //------ Creer le protocole clulink ------
  iCluLinkPos = PROTOCOL_setCluLink(sCluLink, CL(13));

  //Ajouter l'indicatif de l'expediteur
  iCluLinkPos = PROTOCOL_addPascalString(sCluLink, sFromNode, iCluLinkPos);

  //Ajouter le user count
  iCluLinkPos = CLULINK_addShort(sCluLink, &shUserCount, iCluLinkPos);

  //Envoyer le protocole
  if( iToAll == TRUE )
  {
	extern int FirstStream;		//Dfinie dans SWITCH.CPP
	extern int LastStream;		//Dfinie dans SWITCH.CPP

	for(int index = FirstStream; index <= LastStream; index++)
	{
		if( ! (STREAMS_level[index] & LEVEL_cluster ) )
			continue; /* Ce n'est pas un cluster */

		if( HOPS_isExternal(index) )
			continue; /* Link external : ne pas distribuer */

		PROTOCOL_sendProt(StreamNum, PC50, sPavillon, sCluLink,
				  iHops, iCluLinkPos, index, -1);
	}

	/* Mettre a jour le user count dans la table */
	HOPS_updateHops(sFromNode, iHops, PC50);
  }
  else
  {
    PROTOCOL_sendProt(0, PC50, sPavillon, sCluLink, iHops, iCluLinkPos, StreamNum, -1);
  }
}

//--------------------------------------------------------------------------
//------ Transmettre un DB request (PC44) ----------------------------------
//--------------------------------------------------------------------------
//Retourne TRUE si OK et FALSE sinon
int PROTOCOL_dbRequest(int StreamNum, char * szFromNode, char * szToNode, int nUserStream, char * szQualifier, char * szKey, char * szFromUser)
{
  char szPavillon[256];
  char szCluLink[256];
  int  nStream;		/*Numero du port sur lequel router*/
  int  nDummy;          /*Necessaire pour l'appel d'une fonction*/

  if( ! NODE_isNodeConnected(szToNode, &nStream, &nDummy) )
  {
    /*Impossible de trouver le node - prevenir celui qui a fait la demande*/
    if( StreamNum != 0 )
    {
      //Ecrire les protocoles ICI pour une requete distante
    }
    return FALSE;
  }

  /*Pas de ping-pong !!*/
  if( StreamNum == nStream )
    return FALSE;

  /* Protocole PAVILLON */
  sprintf(szPavillon, "PC44^%s^%s^%d^%s^%1s^%s^", szToNode, szFromNode,
	  nUserStream, szQualifier, szKey, szFromUser);

  /* Protocole CLULINK */
  TOOLS_maxLength(szQualifier, 20);
  TOOLS_maxLength(szKey,       20);
  iCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(160));
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromNode,  iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szToNode,    iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromUser,  iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szQualifier, iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szKey,       iCluLinkPos);
  szCluLink[iCluLinkPos++] = (char) nUserStream;

  /*Envoyer le protocole*/
  PROTOCOL_sendProt(0, PC44, szPavillon, szCluLink, 100, iCluLinkPos, nStream, -1);

  return TRUE;
}

//--------------------------------------------------------------------------
//------ Transmettre un DB response (PC45) ---------------------------------
//--------------------------------------------------------------------------
void PROTOCOL_dbResponse(int StreamNum, char * szFromNode, char * szToNode, int nUserStream, char * szInfo)
{
  char szPavillon[256];
  char szCluLink[256];
  int  nStream;		/*Numero du port sur lequel router*/
  int  nDummy;          /*Necessaire pour l'appel d'une fonction*/

  if( ! NODE_isNodeConnected(szToNode, &nStream, &nDummy) )
    return;	/*Impossible de trouver le node !*/

  /*Pas de ping-pong !!*/
  if( StreamNum == nStream )
    return;

  TOOLS_maxLength(szInfo, 128);	/*128 parce que je l'ai decide !*/

  /* Protocole PAVILLON */
  sprintf(szPavillon, "PC45^%s^%s^%d^%1s^", szToNode, szFromNode, nUserStream, szInfo);

  /* Protocole CLULINK */
  iCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(161));
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromNode,  iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szToNode,    iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szInfo,      iCluLinkPos);
  szCluLink[iCluLinkPos++] = (char) nUserStream;

  /* Envoyer le protocole */
  PROTOCOL_sendProt(0, PC45, szPavillon, szCluLink, 100, iCluLinkPos, nStream, -1);
}

//--------------------------------------------------------------------------
//------ Transmettre un DB complete (PC46) ---------------------------------
//--------------------------------------------------------------------------
#pragma argsused
void PROTOCOL_dbComplete(int StreamNum, char * szFromNode, char * szToNode, int nUserStream)
{
  char szPavillon[256];
  char szCluLink[256];
  int  nStream;		/*Numero du port sur lequel router*/
  int  nDummy;          /*Necessaire pour l'appel d'une fonction*/

  if( ! NODE_isNodeConnected(szToNode, &nStream, &nDummy) )
    return;	/*Impossible de trouver le node !*/

  /* Protocole PAVILLON */
  sprintf(szPavillon, "PC46^%s^%s^%d^", szToNode, szFromNode, nUserStream);

  /* Protocole CLULINK */
  iCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(162));
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szFromNode,  iCluLinkPos);
  iCluLinkPos = PROTOCOL_addPascalString(szCluLink, szToNode,    iCluLinkPos);
  szCluLink[iCluLinkPos++] = (char) nUserStream;

  /* Envoyer le protocole */
  PROTOCOL_sendProt(0, PC46, szPavillon, szCluLink, 100, iCluLinkPos, nStream, -1);
}

/*--------------------------------------------------------------------------
  ------ Transmettre PC38 (Connected node list) ----------------------------
  -------------------------------------------------------------------------- */
void PROTOCOL_sendConnectedNodeList(int StreamNum)
{
	/* Ce protocol est transmit inuquement pour compatibilite avec
	   d'autre systemes - Dxnet ne l'utilise pas */

	int	nSize = 0;
	int	nCallSignLength;
	int	nNodeStream;
	int	nNodePos;
	char	szNodeCall[16];

	if( PROTOCOL_linkType[StreamNum] == CLU_LINK )
		return; 	/* Termine ! */

	for(nNodeStream = 0; nNodeStream <= 64; nNodeStream++)
	{
		/* Est-ce un cluster ou MYCALL ? */
		if( nNodeStream == 0 && ! (STREAMS_level[nNodeStream] & LEVEL_cluster) )
			continue; /* Non ! */

		/* Pas de ping-pong ... */
		if( nNodeStream == StreamNum )
			continue;

		/* Parcourir toutes les positions de ce stream */
		for( nNodePos = 0; nNodePos < NODE_MAXANODE; nNodePos++)
		{
			/* Indicatif du cluster */
			NODE_getNodeCall(nNodeStream, nNodePos, szNodeCall);
			nCallSignLength = strlen(szNodeCall);

			if( szNodeCall[0] == SNULL )
				continue;	/* Pas d'indicatif */

			/* Ne pas depasser 240 caracteres */
			if( (nSize + nCallSignLength) > 240 )
			{
				/* Fin de la trame */
				BUFFERS_printBuff(StreamNum, OUT, "^~\n");
				nSize = 0;
			}


			if( nSize == 0 )
				BUFFERS_printBuff(StreamNum, OUT, "PC38^");	/* Debut de la trame */
			else
				BUFFERS_printBuff(StreamNum, OUT, ",");		/* Separation entre champs */

			nSize += nCallSignLength + 6;

			BUFFERS_printBuff(StreamNum, OUT, "%s", szNodeCall);
		}
	}

	if( nSize != 0 )
		BUFFERS_printBuff(StreamNum, OUT, "^~\n");
}

/*------------------------------------------------------------------------
  ------ Transmettre un Remote CMD ---------------------------------------
  ------------------------------------------------------------------------
  Retourne TRUE si le protocole a pu tre rout et qu'il n'y a pas eu
  d'erreur, et FALSE sinon */
int PROTOCOL_rCmd(int StreamNum, char * pszFromPC, char * pszToPC, char * pszCmd)
{
	char szMyNodeCall[16];

	/* Indicatif du serveur */
	NODE_getNodeCall(0, 0, szMyNodeCall);

	/* Ce protocol est il pour moi ? */
	if( ! strcmp(pszToPC, szMyNodeCall) )
	{
		/* OUI : Traiter la commande */
		FILE * 	fPtr;
		char    szBuffer[256];
		char	szResponse[512];

		unlink(TEMP_RCMD_FILE);
		CMD_execute(RCMD_VIRTSTREAM + StreamNum, pszCmd);

		/* Renvoyer la reponse */
		fPtr = fopen(TEMP_RCMD_FILE, "rt");
		if( ! fPtr )
		{
			perror("fopen in PROTOCOL_rCmd");
			return FALSE;
		}

		while( fgets(szBuffer, 255, fPtr) )
		{
			TOOLS_removeNR(szBuffer);
			sprintf(szResponse, "%s: %s", szMyNodeCall, szBuffer);
			PROTOCOL_rCmdResponse(0, pszToPC, pszFromPC, szResponse);
		}

		fclose(fPtr);
	}
	else
	{
		/* NON : faire suivre le protocole */
		int	nStream;
		int	nNodePos;
		char	szPavillon[256];
		char	szCluLink[256];
		int	nCluLinkPos;

		/* Chercher le stream sur lequel router le protocole */
		if( ! NODE_isNodeConnected(pszToPC, &nStream, &nNodePos) )
			return FALSE;

		/* Creer le protocole Pavillon */
		sprintf(szPavillon, "PC34^%s^%s^%s^", pszToPC,
						      pszFromPC,
						      pszCmd);

		/* Creer le protocole CluLink */
		nCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(200));
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszFromPC, nCluLinkPos);
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszToPC, nCluLinkPos);
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszCmd, nCluLinkPos);

		/* Transmettre le protocole */
		PROTOCOL_sendProt(StreamNum, PC34, szPavillon, szCluLink, 100, nCluLinkPos, nStream, -1);
	} /* End IF */

	return TRUE;
}

/*------------------------------------------------------------------------
  ------ Transmettre une reponse a un Remote CMD -------------------------
  ------------------------------------------------------------------------
  Retourne TRUE si le protocole a pu tre rout, et FALSE sinon */
int PROTOCOL_rCmdResponse(int StreamNum, char * pszFromPC, char * pszToPC, char * pszResponse)
{
	char szMyNodeCall[16];

	/* Indicatif du serveur */
	NODE_getNodeCall(0, 0, szMyNodeCall);

	/* Limiter la longueur de la chaine a 200 octets afin d'eviter les
	   effet de bords */
	TOOLS_maxLength(pszResponse, 254);

	/* Ce protocol est il pour moi ? */
	if( ! strcmp(pszToPC, szMyNodeCall) )
	{
		/* OUI : Traiter la commande */
		char szBuffer[256];

		strcpy(szBuffer, pszResponse);
		strcat(szBuffer, "\n");
		USERS_sendAll(szBuffer, USERS_SYSOP, -1);
	}
	else
	{
		/* NON : faire suivre le protocole */
		int	nStream;
		int	nNodePos;
		char	szPavillon[256];
		char	szCluLink[256];
		int	nCluLinkPos;

		/* Chercher le stream sur lequel router le protocole */
		if( ! NODE_isNodeConnected(pszToPC, &nStream, &nNodePos) )
			return FALSE;

		/* Creer le protocole Pavillon */
		sprintf(szPavillon, "PC35^%s^%s^%s^", pszToPC,
						      pszFromPC,
						      pszResponse);

		/* Creer le protocole CluLink */
		nCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(201));
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszFromPC, nCluLinkPos);
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszToPC, nCluLinkPos);
		nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszResponse, nCluLinkPos);

		/* Transmettre le protocole */
		PROTOCOL_sendProt(StreamNum, PC35, szPavillon, szCluLink, 100, nCluLinkPos, nStream, -1);
	} /* End IF */

	return TRUE;

}

/*------------------------------------------------------------------------
  ------ Transmettre un explicit disconnect ------------------------------
  ------------------------------------------------------------------------
  Retourne TRUE si le protocole a pu tre route, et FALSE sinon */
int PROTOCOL_sendExplicitDisconnect(int StreamNum, char * pszToPC, char * pszReason)
{
	char	szMyNodeCall[16];
	int	nStream;
	int	nNodePos;
	char	szPavillon[256];
	char	szCluLink[256];
	int	nCluLinkPos;

	/* Chercher le stream sur lequel router le protocole */
	if( ! NODE_isNodeConnected(pszToPC, &nStream, &nNodePos) )
		return FALSE;

	if( nNodePos != 0 )
		return FALSE; 	/* erreur */

	/* Informer les adjacents */
	PROTOCOL_clusterStreamDisconnected(nStream, pszReason, -1);
//	PROTOCOL_nodeDisconnect(nStream, pszToPC, pszReason, 1, 100);

	/* Prepaper et envoyer un PC39 */
	NODE_getNodeCall(0, 0, szMyNodeCall);

	/* Creer le protocole Pavillon */
	sprintf(szPavillon, "PC39^%s^%s^", szMyNodeCall, pszReason);

	/* Creer le protocole CluLink */
	nCluLinkPos = PROTOCOL_setCluLink(szCluLink, CL(5));
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, szMyNodeCall, nCluLinkPos);
	nCluLinkPos = PROTOCOL_addPascalString(szCluLink, pszToPC,      nCluLinkPos);

	/* Transmettre le protocole */
	PROTOCOL_sendProt(StreamNum, PC39, szPavillon, szCluLink, 100, nCluLinkPos, nStream, -1);

	return TRUE;
}

/*------------------------------------------------------------------------
  ------ Traiter un un explicit disconnect recu --------------------------
  ------------------------------------------------------------------------
  Retourne TRUE si le protocole a pu tre route, et FALSE sinon */
int PROTOCOL_rcvExplicitDisconnect(int StreamNum, char * pszFromPC, char * pszReason, int nHops)
{
	extern char	STREAMS_level[MAX_STREAMS];
	extern char	STREAMS_callsign[MAX_STREAMS][10];
	int	nStream;
	int	nNodePos;

	/* Chercher le stream sur lequel router le protocole */
	if( ! NODE_isNodeConnected(pszFromPC, &nStream, &nNodePos) )
		return FALSE;

	/* Verifications */
	if( nStream != StreamNum || nNodePos != 0 )
	{
		/* Erreur : deconnection hard */
		extern unsigned long 	COMMANDS_statut[MAX_STREAMS];
		COMMANDS_statut[StreamNum] = COMMANDS_statutDisconnect;
		return FALSE;
	}

	/* Prevenir les adjacents */
	PROTOCOL_clusterStreamDisconnected(StreamNum, pszReason, nHops);
#if 0
	for(int index = 0; index < NODE_MAXANODE; index++)
	{
		char szNodeCall[16];

		NODE_getNodeCall(StreamNum, index, szNodeCall);
		if( szNodeCall[0] == SNULL )
			continue;

		NODE_deleteNode(StreamNum, szNodeCall);
		BUFFERS_printBuff(BUFFER_SCREEN1, OUT,
				  "Node logged out : %s\n", szNodeCall);

		/* NODE_deleteNode dcalant d'un cran les nodes,
		   decrementer l'index */
		index--;

		/* Informer les adjacent */
		PROTOCOL_nodeDisconnect(StreamNum, szNodeCall, pszReason, 1, nHops);
	}
#endif

	/* Deconnecter ce cluster */
	STREAMS_level[nStream] = LEVEL_discByOp;
	STREAMS_disconnect(nStream);
	SWITCH_discSwitch (nStream);
	STREAMS_callsign[nStream][0] = SNULL;

	return TRUE;
}

/*------------------------------------------------------------------------
  ------ Informer les adjacents de la deconnexion qu'un stream cluster ---
  ------ et liberer la memoire que pour ce stream ------------------------
  ------------------------------------------------------------------------
  StreamNum : stream du cluster deconnecte
  nHops : Hops du PC21 a transmetre
  ------------------------------------------------------------------------
  La fonction retourne TRUE si OK et FALSE sinon */
int PROTOCOL_clusterStreamDisconnected(int StreamNum, char * pszReason, int nHops)
{
	extern char STREAMS_level[MAX_STREAMS]; // + RCMD_VIRTSTREAM];
	char szNodeCall[16];
	int  hops;

	/* Verifier que le stream est bien celui d'un cluster */
	if( ! (STREAMS_level[StreamNum] & LEVEL_cluster) )
		return FALSE;

	for(int index = 0; index < NODE_MAXANODE; index++)
	{
		NODE_getNodeCall(StreamNum, index, szNodeCall);
		if( szNodeCall[0] == SNULL )
			continue;

		NODE_deleteNode(StreamNum, szNodeCall);
		BUFFERS_printBuff(BUFFER_SCREEN1, OUT,
				  "Node logged out : %s\n", szNodeCall);

		/* NODE_deleteNode dcalant d'un cran les nodes,
		   decrementer l'index */
		index--;

		/* Informer les adjacent */
		if (nHops == -1 )
		{
			/* Comme c'est MOI qui demande la deconnexion, il
			   faut determiner quel hops je vais devoir
			   utiliser dans le PC21 pour ce cluster */
			int nPc16, nPc19, nPc50;
			HOPS_getHops(szNodeCall, 0, nPc16, nPc19, nPc50);
			hops = nPc19;
			hops = 100;
		}
		else
			hops = nHops;

		PROTOCOL_nodeDisconnect(StreamNum, szNodeCall, pszReason, 1, hops);
	}

	return TRUE;
}
